import functools
import re
from .IosSectionParsers import IosConfigLine
from net_models.models.interfaces.InterfaceCommon import *
from net_models.models.interfaces.InterfaceModels import *
from net_models.models.interfaces.L2InterfaceModels import *
from net_models.models.interfaces.L3InterfaceModels import *

from net_models.validators import expand_vlan_range

from typing import (
    Union
)

from net_parser.utils import compile_regex, re_search, re_search_lines, re_filter_lines

INTERFACE_SECTION_REGEX = re.compile(pattern=r'^interface \S+$', flags=re.MULTILINE)


class IosInterfaceParser(IosConfigLine, regex=INTERFACE_SECTION_REGEX):

    _name_regex = re.compile(pattern=r"^interface (?P<name>.*)\Z", flags=re.MULTILINE)
    _description_regex = re.compile(pattern=r"^ description (?P<description>.*?)\Z", flags=re.MULTILINE)
    _ipv4_addr_regex = re.compile(pattern=r"^ ip address (?P<address>(?:\d{1,3}\.){3}\d{1,3}) (?P<mask>(?:\d{1,3}\.){3}\d{1,3})(?: (?P<secondary>secondary))?", flags=re.MULTILINE)
    _vrf_regex = re.compile(pattern=r"^(?:\sip)?\svrf\sforwarding\s(?P<vrf>\S+)", flags=re.MULTILINE)
    _shutdown_regex = re.compile(pattern=r"^ (?P<shutdown>shutdown)\Z", flags=re.MULTILINE)
    _no_shutdown_regex = re.compile(pattern=r"^ (?P<no_shutdown>no shutdown)\Z", flags=re.MULTILINE)
    _cdp_regex = re.compile(pattern=r"^ (?P<cdp_enabled>cdp enable)", flags=re.MULTILINE)
    _no_cdp_regex = re.compile(pattern=r"^ (?P<no_cdp_enabled>no cdp enable)", flags=re.MULTILINE)
    _lldp_transmit_regex = re.compile(pattern=r"^ lldp transmit", flags=re.MULTILINE)
    _no_lldp_transmit_regex = re.compile(pattern=r"^ no lldp transmit", flags=re.MULTILINE)
    _lldp_receive_regex = re.compile(pattern=r"^ lldp receive", flags=re.MULTILINE)
    _no_lldp_receive_regex = re.compile(pattern=r"^ no lldp receive", flags=re.MULTILINE)
    _mtu_regex = re.compile(pattern=r"r^ mtu (?P<mtu>\d+)", flags=re.MULTILINE)
    _ip_mtu_regex = re.compile(pattern=r"^ ip mtu (?P<ip_mtu>\d+)", flags=re.MULTILINE)
    _bandwidth_regex = re.compile(pattern=r"^ bandwidth (?P<bandwidth>\d+)", flags=re.MULTILINE)
    _delay_regex = re.compile(pattern=r"^ delay (?P<delay>\d+)", flags=re.MULTILINE)
    _load_interval_regex = re.compile(pattern=r"^ load-interval (?P<load_interval>\d+)")


    _ospf_process_regex = re.compile(pattern=r"^ ip ospf (?P<process_id>\d+) area (?P<area>\d+)$", flags=re.MULTILINE)
    _ospf_network_type_regex = re.compile(pattern=r"^ ip ospf network (?P<network_type>\S+)", flags=re.MULTILINE)
    _ospf_priority_regex = re.compile(pattern=r"^ ip ospf priority (?P<priority>\d+)", flags=re.MULTILINE)
    _ospf_cost_regex = re.compile(pattern=r"^ ip ospf cost (?P<cost>\d+)", flags=re.MULTILINE)
    _ospf_bfd_regex = re.compile(pattern=r"^ ip ospf bfd(?: (?:(?P<disable>disable)|(?P<strict_mode>strict-mode)))?", flags=re.MULTILINE)
    _ospf_timers = re.compile(pattern=r"^ ip ospf (?P<timer>\S+?)-interval (?P<interval>\d+)\Z", flags=re.MULTILINE)
    _ospf_authentication_method = re.compile(pattern=r"^ ip ospf authentication (?P<method>\S+)(?: (?P<keychain>\S+))?", flags=re.MULTILINE)
    _ospf_authentication_key = re.compile(pattern=r"^ ip ospf authentication-key(?: (?P<encryption_type>\d))? (?P<value>\S+)", flags=re.MULTILINE)


    _isis_process_regex = re.compile(pattern=r"^ ip router isis (?P<name>\S+)\Z")
    _isis_network_type_regex = re.compile(pattern=r"^ isis network (?P<network_type>\S+)", flags=re.MULTILINE)
    _isis_circuit_type_regex = re.compile(pattern=r"^ isis circuit-type (?P<circuit_type>\S+)", flags=re.MULTILINE)
    _isis_metric_regex = re.compile(pattern=r"^ isis metric (?P<metric>\d+) (?P<level>\S+)", flags=re.MULTILINE)


    _native_vlan_regex = re.compile(pattern=r"^ switchport trunk native vlan (?P<native_vlan>\d+)", flags=re.MULTILINE)
    _trunk_encapsulation_regex = re.compile(pattern=r"^ switchport trunk encapsulation (?P<encapsulation>dot1q|isl|negotiate)", flags=re.MULTILINE)
    _switchport_mode_regex = re.compile(pattern=r"^ switchport mode (?P<switchport_mode>access|trunk|dot1q-tunnel|private-vlan|dynamic)")
    _switchport_nonegotiate_regex = re.compile(pattern=r"^ switchport nonegotiate")
    _trunk_allowed_vlans_regex = re.compile(pattern=r"^ switchport trunk allowed vlan(?: add)? (?P<allowed_vlans>\S+)", flags=re.MULTILINE)
    _access_vlan_regex = re.compile(pattern=r"^ switchport access vlan (?P<access_vlan>\d+)", flags=re.MULTILINE)
    _voice_vlan_regex = re.compile(pattern=r"^ switchport voice vlan (?P<voice_vlan>\d+)")

    _hsrp_version_regex = re.compile(pattern=r"^ standby version (?P<version>\d)", flags=re.MULTILINE)
    _hsrp_ipv4_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) ip (?P<address>\S+)(?: (?P<secondary>secondary))?", flags=re.MULTILINE)
    _hsrp_priority_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) priority (?P<priority>\S+)", flags=re.MULTILINE)
    _hsrp_preemption_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) (?P<preemption>preempt)", flags=re.MULTILINE)
    _hsrp_timers_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) timers (?:(?P<milliseconds>msec) )?(?P<hello>\d+) (?P<hold>\d+)", flags=re.MULTILINE)
    _hsrp_track_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) track (?P<track_id>\d+) (?P<action>shutdown|decrement)(?: (?P<decrement_value>\d+))?", flags=re.MULTILINE)
    _hsrp_authentication_regexes = [
        re.compile(pattern=r"^ standby (?P<group_id>\d+) authentication (?:(?P<method>md5|text) )?(?:key-string )?(?:(?P<encryption_type>\d) )?(?P<value>\S+)$", flags=re.MULTILINE),
        re.compile(pattern=r"^ standby (?P<group_id>\d+) authentication md5 (?P<method>key-chain) (?P<keychain>\S+)$", flags=re.MULTILINE)
    ]
    _hsrp_name_regex = re.compile(pattern=r"^ standby (?P<group_id>\d+) name (?P<name>\S+)$", flags=re.MULTILINE)

    _proxy_arp_regex = re.compile(pattern=r"^ (?:(?P<no>no) )?ip proxy-arp$", flags=re.MULTILINE)

    _service_policy_regex = re.compile(pattern=r"^ service-policy (?P<direction>input|output) (?P<name>\S+$)", flags=re.MULTILINE)




    def __init__(self, number: int, text: str, config, verbosity: int = 4):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name="IosInterfaceLine")

    @property
    def get_type(self):
        types = super().get_type
        types.append('interface')
        return types


    @functools.cached_property
    def name(self) -> Union[str, None]:
        return self.re_match(regex=self._name_regex, group=1)

    @functools.cached_property
    def description(self) -> Union[str, None]:
        candidates = self.re_search_children(regex=self._description_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates)

    @functools.cached_property
    def is_enabled(self) -> Union[bool, None]:
        shutdown_candidates = self.re_search_children(regex=self._shutdown_regex, group='shutdown')
        no_shutdown_candidates = self.re_search_children(regex=self._no_shutdown_regex, group='no_shutdown')
        if len(shutdown_candidates) == 1:
            return False
        elif len(no_shutdown_candidates):
            return True
        else:
            if self.config.DEFAULTS.INTERFACES_DEFAULT_NO_SHUTDOWN is not None:
                self.logger.info(msg=f"Interface {self.name}: Using platform default value for interface admin state.")
                return self.config.DEFAULTS.INTERFACES_DEFAULT_NO_SHUTDOWN
            else:
                self.logger.debug(msg=f"Interface {self.name}: Platform default for interface admin state not set.")
                return None

    @functools.cached_property
    def is_routed(self):
        candidates = self.re_search_children(regex=re.compile(pattern="(?:no )?ip address"))
        self.logger.debug(msg=f"Interface {self.name}: {candidates=}")
        if len(candidates):
            return True
        else:
            return False

    @functools.cached_property
    def is_switched(self):
        candidates = self.re_search_children(regex=re.compile(pattern="(?:no )?ip address"))
        self.logger.debug(msg=f"Interface {self.name}: {candidates=}")
        if len(candidates):
            return False
        else:
            return True



    @functools.cached_property
    def cdp(self) -> Union[InterfaceCdpConfig, None]:
        cdp_candidates = self.re_search_children(regex=self._cdp_regex, group=1)
        no_cdp_candidates = self.re_search_children(regex=self._no_cdp_regex, group=1)
        if len(cdp_candidates) == 1:
            return InterfaceCdpConfig(enabled=True)
        elif len(no_cdp_candidates):
            return InterfaceCdpConfig(enabled=True)
        else:
            if self.config.DEFAULTS.INTERFACES_DEFAULT_CDP_ENABLED is None:
                self.logger.debug("Platform default for CDP not set.")
                return None
            else:
                self.logger.warning(msg="Using platform default value for interface CDP.")
                return InterfaceCdpConfig(enabled=self.config.DEFAULTS.INTERFACES_DEFAULT_CDP_ENABLED)


    @functools.cached_property
    def lldp(self) -> Union[InterfaceLldpConfig, None]:
        lldp_transmit_candidates = self.re_search_children(regex=self._lldp_transmit_regex)
        no_lldp_transmit_candidates = self.re_search_children(regex=self._no_lldp_transmit_regex)
        lldp_receive_candidates = self.re_search_children(regex=self._lldp_receive_regex)
        no_lldp_receive_candidates = self.re_search_children(regex=self._no_lldp_receive_regex)
        data = {}

        if len(lldp_transmit_candidates):
            data["transmit"] = True
        elif len(no_lldp_transmit_candidates):
            data["transmit"] = False
        elif self.config.DEFAULTS.INTERFACES_DEFAULT_LLDP_ENABLED is not None:
            self.logger.warning(msg="Using platform default value for interface LLDP transmit.")
            data["transmit"] = self.config.DEFAULTS.INTERFACES_DEFAULT_LLDP_ENABLED

        if len(lldp_receive_candidates):
            data["receive"] = True
        elif len(no_lldp_receive_candidates):
            data["receive"] = False
        elif self.config.DEFAULTS.INTERFACES_DEFAULT_LLDP_ENABLED is not None:
            self.logger.warning(msg="Using platform default value for interface LLDP receive.")
            data["receive"] = self.config.DEFAULTS.INTERFACES_DEFAULT_LLDP_ENABLED

        if any(data.values()):
            return InterfaceLldpConfig(**data)
        else:
            self.logger.debug("Platform default for LLDP not set.")
            return None

    @functools.cached_property
    def mtu(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._mtu_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def ip_mtu(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._ip_mtu_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def bandwidth(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._bandwidth_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def delay(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._delay_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def load_interval(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._load_interval_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def vrf(self) -> Union[str, None]:
        candidates = self.re_search_children(regex=self._vrf_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates)

    @functools.cached_property
    def service_policy(self) -> Union[InterfaceServicePolicy, None]:
        candidates = self.re_search_children(regex=self._service_policy_regex, group='ALL')
        if len(candidates) == 0:
            return None

        data = {x['direction']: x['name'] for x in candidates}
        model = InterfaceServicePolicy.parse_obj(data)
        return model


    @functools.cached_property
    def ipv4_addresses(self) -> Union[InterfaceIPv4Container, None]:
        candidates = self.re_search_children(regex=self._ipv4_addr_regex, group='ALL')
        if len(candidates) == 0:
            return None
        else:
            candidates = [self._val_to_bool(entry=x, keys=['secondary']) for x in candidates]
            candidates = [{'address': f"{x['address']}/{x['mask']}", 'secondary': x['secondary']} for x in candidates]
            return InterfaceIPv4Container(addresses=[InterfaceIPv4Address(**x) for x in candidates])

    @functools.cached_property
    def hsrp(self):
        unprocessed_lines = self.re_search_children(regex=self.compile_regex(pattern=r"^ standby .*"))
        if not len(unprocessed_lines):
            return None
        data = InterfaceHsrp(version=1)
        version, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_version_regex, group='version', pop_matches=True)
        version = self.first_candidate_or_none(candidates=version, wanted_type=int)
        if version is not None:
            data.version = version
        ip_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_ipv4_regex, group='ALL', pop_matches=True)
        ip_candidates = [self._val_to_bool(entry=x, keys=['secondary']) for x in ip_candidates]
        # print(f"{ip_candidates=}")
        priority_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_priority_regex, group="ALL")
        # print(f"{priority_candidates=}")
        preemtion_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_preemption_regex, group="ALL")
        preemtion_candidates = [self._val_to_bool(entry=x, keys=['preemption']) for x in preemtion_candidates]
        # print(f"{preemtion_candidates=}")
        timers_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_timers_regex, group="ALL")
        timers_candidates = [self._val_to_bool(entry=x, keys=['milliseconds']) for x in timers_candidates]
        # print(f"{timers_candidates=}")
        track_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_track_regex, group="ALL")
        # print(f"{track_candidates=}")
        authentication_candidates = self.re_search_children_multipattern(regexes=self._hsrp_authentication_regexes, group="ALL")
        # print(f"{authentication_candidates=}")
        name_candidates, unprocessed_lines = re_filter_lines(lines=unprocessed_lines, regex=self._hsrp_name_regex, group="ALL")
        # print(f"{name_candidates=}")


        # Get list of all group_ids
        group_ids = set()
        for candidates in [
            ip_candidates, priority_candidates, preemtion_candidates, timers_candidates, track_candidates,
            authentication_candidates, name_candidates
        ]:
            group_ids = group_ids | set([x['group_id'] for x in candidates])
        group_ids = sorted(list(group_ids), key=lambda x: int(x))

        # Populate Group Models
        for group_id in group_ids:
            group = InterfaceHsrpGroup(group_id=group_id)
            for c in [x for x in ip_candidates if x['group_id'] == group_id]:
                if group.ipv4 is None:
                    group.ipv4 = []
                include_keys = {k:v for k,v in c.items() if k in HsrpIpv4Address.__fields__.keys()}
                group.ipv4.append(HsrpIpv4Address.parse_obj({k:v for k,v in c.items() if k in include_keys }))
            # Priority
            for c in [x for x in priority_candidates if x['group_id'] == group_id]:
                group.priority = c['priority']
            # Preemption
            for c in [x for x in preemtion_candidates if x['group_id'] == group_id]:
                group.preemption = c['preemption']
            # Timers
            for c in [x for x in timers_candidates if x['group_id'] == group_id]:
                include_keys = {k:v for k,v in c.items() if k in HsrpTimers.__fields__.keys()}
                group.timers = HsrpTimers.parse_obj({k:v for k,v in c.items() if k in include_keys})
            # Tracks
            for c in [x for x in track_candidates if x['group_id'] == group_id]:
                include_keys = {k:v for k,v in c.items() if k in HsrpTrack.__fields__.keys()}
                if group.tracks is None:
                    group.tracks = []
                group.tracks.append(HsrpTrack.parse_obj({k:v for k,v in c.items() if k in include_keys and v is not None}))
            # Authentication
            for c in [x for x in authentication_candidates if x['group_id'] == group_id]:
                authentication = None
                if c['method'] is None:
                    c['method'] = 'text'
                if c['method'] in ['text', 'md5']:
                    include_keys = {k:v for k,v in c.items() if k in KeyBase.__fields__.keys()}
                    key = KeyBase.parse_obj({k:v for k,v in c.items() if k in include_keys and v is not None})
                    authentication = HsrpAuthentication(method=c['method'], key=key)
                elif c['method'] == 'key-chain':
                    authentication = HsrpAuthentication(method=c['method'], keychain=c['keychain'])
                group.authentication = authentication
            # Name
            for c in [x for x in name_candidates if x['group_id'] == group_id]:
                group.name = c['name']
            if data.groups is None:
                data.groups = []
            data.groups.append(group)

        if len(unprocessed_lines):
            pass
            # data.extra_config = [x.text for x in unprocessed_lines]
        data = data.copy()
        return data

    @functools.cached_property
    def switchport_mode(self):
        candidates = self.re_search_children(regex=self._switchport_mode_regex, group='switchport_mode')
        return self.first_candidate_or_none(candidates=candidates)

    @functools.cached_property
    def switchport_negotiation(self) -> bool:
        """
        Returns boolean representation whether negotiation (such as DTP) is enabled (True) or disabled (False)
        Returns: bool

        """
        candidates = self.re_search_children(regex=self._switchport_nonegotiate_regex)
        if self.first_candidate_or_none(candidates=candidates) is None:
            return True # Negotiation is NOT disabled
        else:
            return False # Negotiation is disabled


    @functools.cached_property
    def access_vlan(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._access_vlan_regex, group='access_vlan')
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def voice_vlan(self) -> Union[int, None]:
        candidates = self.re_search_children(regex=self._voice_vlan_regex, group='voice_vlan')
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def trunk_allowed_vlans(self) -> Union[List[int], Literal['all', 'none'], None]:
        candidates = self.re_search_children(regex=self._trunk_allowed_vlans_regex, group='allowed_vlans')
        print(candidates)
        allowed_vlans = None
        if not len(candidates):
            pass
        else:
            if len(candidates) == 1:
                if candidates[0] in ['all', 'none']:
                    return candidates[0]
            allowed_vlans = []
            for candidate in candidates:
                allowed_vlans.extend(candidate.split(','))
            allowed_vlans = expand_vlan_range(vlan_range=allowed_vlans)
        return allowed_vlans

    @functools.cached_property
    def proxy_arp_enabled(self) -> bool:
        candidates = self.re_search_children(regex=self._proxy_arp_regex, group='ALL')
        candidate = self.first_candidate_or_none(candidates=candidates)

        if candidate is not None:
            candidate = self._val_to_bool(entry=candidate, keys=['no'])
            if candidate['no'] is True:
                # no ip proxy-arp
                return False
            elif candidate['no'] is False:
                # ip proxy-arp
                return True
        else:
            self.logger.debug(msg=f"Interface {self.name}: Using global value for Proxy ARP: {self.config.proxy_arp_enabled}")
            # If not specified, return global value
            return self.config.proxy_arp_enabled

    @functools.cached_property
    def ospf(self) -> Union[InterfaceOspfConfig, None]:
        data = {}
        process_candidates = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_process_regex, group='ALL'))
        network_type = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_network_type_regex, group='ALL'))
        cost = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_cost_regex, group='ALL'))
        priority = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_cost_regex, group='ALL'))
        bfd = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_bfd_regex, group="ALL"))
        timers = self.re_search_children(regex=self._ospf_timers, group="ALL")
        authentication_method = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_authentication_method, group="ALL"))
        authentication_key = self.first_candidate_or_none(self.re_search_children(regex=self._ospf_authentication_key, group="ALL"))

        if process_candidates is not None:
            data.update(process_candidates)
        if network_type is not None:
            data.update(network_type)
        if cost is not None:
            data.update(cost)
        if priority is not None:
            data.update(priority)
        # BFD
        if bfd is not None:
            bfd = self._val_to_bool(entry=bfd, keys=['disable', 'strict_mode'])
            if bfd['disable']:
                data.update({"bfd": False})
            elif bfd['strict_mode']:
                data.update({"bfd": 'strict-mode'})
            else:
                data.update({"bfd": True})
        # Timers
        if len(timers):
            timers = InterfaceOspfTimers(**{x["timer"]:x['interval'] for x in timers})
            data['timers'] = timers
        # Authentication
        if any([authentication_key, authentication_method]):
            authentication_data = {}
            if authentication_method:
                authentication_data.update({k:v for k, v in authentication_method.items() if v is not None})
            if authentication_key:
                if authentication_key['encryption_type'] is None:
                    authentication_key['encryption_type'] = 0
                authentication_data.update({'key': authentication_key})
            authentication = InterfaceOspfAuthentication(**authentication_data)
            data['authentication'] = authentication
        # Convert to model
        if len(data) == 0:
            return None
        else:
            return InterfaceOspfConfig(**data)


    @functools.cached_property
    def isis(self):
        # TODO: Complete
        data = {}
        patterns = [
            self._isis_process_regex,
            self._isis_network_type_regex,
            self._isis_circuit_type_regex
        ]
        results = self.re_search_children_multipattern(regexes=patterns, group="ALL")
        for entry in results:
            data.update(entry)
        results = self.re_search_children(regex=self._isis_metric_regex, group="ALL")
        if len(results):
            data['metric'] = list(results)



    def to_model(self) -> InterfaceModel:
        self.logger.debug(msg=f"Building model for interface {self.name}")
        model = InterfaceModel(
            name=self.name,
        )
        if self.is_enabled is not None:
            model.enabled = self.is_enabled

        if self.description is not None:
            model.description = self.description


        l3_parameters = [
            self.ipv4_addresses,
            self.ip_mtu,
            self.vrf,
            self.ospf,
            self.isis
        ]
        if any(l3_parameters):
            if model.l3_port is None:
                model.l3_port = InterfaceRouteportModel()

        discovery_protocols = [
            self.cdp,
            self.lldp
        ]
        if any(discovery_protocols):
            self.logger.debug(msg=f"Discovery Protocols: {discovery_protocols}")
            if model.discovery_protocols is None:
                model.discovery_protocols = InterfaceDiscoveryProtocols()


        if self.load_interval is not None:
            model.load_interval = self.load_interval
        if self.delay is not None:
            model.delay = self.delay
        if self.bandwidth is not None:
            model.bandwidth = self.bandwidth

        if self.cdp is not None:
            model.discovery_protocols.cdp = self.cdp

        if self.lldp is not None:
            model.discovery_protocols.lldp = self.lldp

        if self.mtu is not None:
            model.mtu = self.mtu

        if self.service_policy is not None:
            model.service_policy = self.service_policy


        if self.ipv4_addresses is not None:
            model.l3_port.ipv4 = self.ipv4_addresses
        if self.vrf is not None:
            model.l3_port.vrf = self.vrf
        if self.ip_mtu is not None:
            model.l3_port.ip_mtu = self.ip_mtu


        # Dynamic Routing Protocols
        if self.ospf is not None:
            model.l3_port.ospf = self.ospf

        if self.isis is not None:
            model.l3_port.isis = self.isis

        if self.hsrp is not None:
            model.l3_port.hsrp = self.hsrp


        return model





import pathlib
import re
import unittest
from tests import BaseNetParserTest


from net_models.models.interfaces.InterfaceCommon import *
from net_models.models.interfaces.L3InterfaceModels import *
from net_models.models.services.ServerModels import *

from net_parser.config import BaseConfigParser, IosConfigParser, BaseConfigLine
VERBOSITY = 5

class TestIosConfigParserLoading(BaseNetParserTest):

    VENDOR = 'ios'
    TEST_CLASS = IosConfigParser

    def test_load_path(self):
        path = self.RESOURCES_DIR.joinpath(self.VENDOR).joinpath('data').joinpath('test_load_01.txt')
        config = self.TEST_CLASS(config=path, verbosity=VERBOSITY)
        config.parse()

    def test_load_string_path(self):
        path = self.RESOURCES_DIR.joinpath(self.VENDOR).joinpath('data').joinpath('test_load_01.txt')
        config = self.TEST_CLASS(config=str(path), verbosity=VERBOSITY)
        config.parse()

    def test_load_string_config(self):
        path = self.RESOURCES_DIR.joinpath(self.VENDOR).joinpath('data').joinpath('test_load_01.txt')
        config = self.TEST_CLASS(config=path.read_text())
        config.parse()

    def test_load_string_list(self):
        path = self.RESOURCES_DIR.joinpath(self.VENDOR).joinpath('data').joinpath('test_load_01.txt')
        config = self.TEST_CLASS(config=path.read_text().split('\n'), verbosity=VERBOSITY)
        config.parse()

    def test_load_single_line_config(self):
        config = self.TEST_CLASS(config="! This might be a config but not a path")
        config.parse()

    def test_load_nonexistent_path(self):
        path = pathlib.Path('/path/does/not/exist.txt')
        config = self.TEST_CLASS(config=path, verbosity=VERBOSITY)
        with self.assertRaises(FileNotFoundError):
            config.parse()


class TestIosConfigParser(BaseNetParserTest):

    TEST_CLASS = IosConfigParser
    VENDOR = "ios"

    def get_config(self) -> IosConfigParser:
        return self.TEST_CLASS(
            self.RESOURCES_DIR.joinpath(self.VENDOR).joinpath('data').joinpath('test_load_01.txt'),
            verbosity=VERBOSITY
        )

    def test_find_objects_01(self):
        config = self.get_config()
        config.parse()
        candidates = config.find_objects(regex=r'^hostname (?P<hostname>.*)')
        self.assertIsInstance(candidates, list)
        self.assertEqual(len(candidates), 1)
        self.assertIsInstance(candidates[0], BaseConfigLine)

    def test_find_objects_01(self):
        config = self.get_config()
        config.parse()
        candidates = config.find_objects(regex=r'^hostname (?P<hostname>.*)', group=1)
        self.assertIsInstance(candidates, list)
        self.assertEqual(len(candidates), 1)
        self.assertIsInstance(candidates[0], str)

    def test_hostname(self):
        config = self.get_config()
        config.parse()
        self.assertIsInstance(config.hostname, str)

    def test_interfaces(self):
        config = self.get_config()
        config.parse()

    def test_interface_01(self):
        data_path, results_path = self.get_test_resources(test_name='interface_01')
        config = self.TEST_CLASS(config=data_path, verbosity=VERBOSITY)
        config.parse()
        interface_lines = list(config.interface_lines)
        for i in interface_lines:
            i.isis
        interfaces_models = list(config.interfaces)

    def test_global_proxy_arp(self):
        default_config = self.TEST_CLASS(config=[])
        default_config.parse()
        full_default_config = self.TEST_CLASS(config=["no ip arp proxy disable"])
        full_default_config.parse()
        disabled_config = self.TEST_CLASS(config=["ip arp proxy disable"])
        disabled_config.parse()
        self.assertEqual(default_config.proxy_arp_enabled, True)
        self.assertEqual(full_default_config.proxy_arp_enabled, True)
        self.assertEqual(disabled_config.proxy_arp_enabled, False)

    def test_ntp(self):
        data_path, results_path = self.get_test_resources(test_name='ntp-01')
        config = self.TEST_CLASS(config=data_path, verbosity=VERBOSITY)
        config.parse()
        want = NtpConfig.parse_obj(self.load_resource_yaml(path=results_path))
        have = config.ntp
        self.assertEqual(want, have)


    def test_logging(self):
        data_path, results_path = self.get_test_resources(test_name='logging-01')
        config = self.TEST_CLASS(config=data_path, verbosity=VERBOSITY)
        config.parse()
        want = LoggingConfig.parse_obj(self.load_resource_yaml(path=results_path))
        have = config.logging
        print(have.yaml(exclude_none=True))
        self.assertEqual(want, have)




class TestIosInterfaceParser(BaseNetParserTest):

    VENDOR = "ios"
    TEST_CLASS = IosConfigParser

    def test_ospf_01(self):
        data_path, results_path = self.get_test_resources(test_name='interface_ospf_01')
        config = IosConfigParser(config=data_path, verbosity=VERBOSITY)
        config.parse()
        want = self.load_resource_yaml(path=results_path)
        have = [x.serial_dict(exclude_none=True) for x in config.interfaces]
        self.assertEqual(want, have)

    def test_is_routed_or_switched(self):
        config_lines = [
            "interface Loopback0",
            " no ip address",
            "interface Vlan100",
            " ip address 192.168.100.1 255.255.255.0",

        ]
        want = {
            "Loopback0": True,
            "Vlan100": True
        }
        config = IosConfigParser(config=config_lines, verbosity=VERBOSITY)
        config.parse()
        for interface_name, result in want.items():
            interface = config.get_interface_line(interface_name=interface_name)
            self.assertEqual(interface.is_routed, result)
            self.assertEqual(interface.is_switched, not result)

    def test_switchport_mode(self):
        config_lines = [
            "interface GigabitEthernet1/0/1",
            " switchport mode access",
            "interface TenGigabitEthernet1/0/1",
            " switchport mode trunk",
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        access_interface = [x for x in config.interface_lines if x.name == "GigabitEthernet1/0/1"][0]
        trunk_interface = [x for x in config.interface_lines if x.name == "TenGigabitEthernet1/0/1"][0]
        self.assertEqual(access_interface.switchport_mode, 'access')
        self.assertEqual(trunk_interface.switchport_mode, 'trunk')

    def test_switchport_nonegotiate(self):
        config_lines = [
            "interface GigabitEthernet1/0/1",
            " switchport mode access",
            "interface TenGigabitEthernet1/0/1",
            " switchport mode trunk",
            " switchport nonegotiate"
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        access_interface = [x for x in config.interface_lines if x.name == "GigabitEthernet1/0/1"][0]
        trunk_interface = [x for x in config.interface_lines if x.name == "TenGigabitEthernet1/0/1"][0]
        self.assertEqual(access_interface.switchport_negotiation, True)
        self.assertEqual(trunk_interface.switchport_negotiation, False)

    def test_access_vlan(self):
        config_lines = [
            "interface GigabitEthernet1/0/1",
            " switchport access vlan 10"
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        interface_line = [x for x in config.interface_lines if x.name == "GigabitEthernet1/0/1"][0]
        self.assertEqual(interface_line.access_vlan, 10)

    def test_voice_vlan(self):
        config_lines = [
            "interface GigabitEthernet1/0/1",
            " switchport voice vlan 10"
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        interface_line = [x for x in config.interface_lines if x.name == "GigabitEthernet1/0/1"][0]
        self.assertEqual(interface_line.voice_vlan, 10)

    def test_trunk_allowed_vlans(self):
        config_lines = [
            "interface Port-channel1",
            " switchport trunk allowed vlan 10-20,21,22",
            " switchport trunk allowed vlan add 30-35,40-50"
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        interface_line = [x for x in config.interface_lines if x.name == "Port-channel1"][0]
        want = [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 30, 31, 32, 33, 34, 35, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50]
        have = interface_line.trunk_allowed_vlans
        self.assertEqual(want, have)

    def test_service_policy(self):
        config_lines = [
            "interface Port-channel1",
            " service-policy input PM-IN",
            " service-policy output PM-OUT",
        ]
        config = IosConfigParser(config=config_lines)
        config.parse()
        interface_line = [x for x in config.interface_lines if x.name == "Port-channel1"][0]
        want = InterfaceServicePolicy(input='PM-IN', output='PM-OUT')
        have = interface_line.service_policy
        self.assertEqual(want, have)


    def test_hsrp(self):
        data_path, results_path = self.get_test_resources(test_name='interface_hsrp_01')
        config = IosConfigParser(config=data_path, verbosity=VERBOSITY)
        config.parse()

        want = self.load_resource_yaml(path=results_path)
        want = InterfaceHsrp.parse_obj(want)
        have = config.get_interface_line(interface_name="Vlan1").hsrp
        self.assertEqual(want, have)


    def test_proxy_arp(self):
        config_1 = IosConfigParser(config=["interface Vlan1"], verbosity=VERBOSITY)
        config_1.parse()
        interface = config_1.get_interface_line(interface_name="Vlan1")
        with self.subTest(msg="All Default"):
            self.assertEqual(interface.proxy_arp_enabled, True)
        config_2 = IosConfigParser(config=["interface Vlan1", " no ip proxy-arp"], verbosity=VERBOSITY)
        config_2.parse()
        interface = config_2.get_interface_line(interface_name="Vlan1")
        with self.subTest(msg="Interface Disabled"):
            self.assertEqual(interface.proxy_arp_enabled, False)
        config_3 = IosConfigParser(config=["ip arp proxy disable", "interface Vlan1",], verbosity=VERBOSITY)
        config_3.parse()
        interface = config_3.get_interface_line(interface_name="Vlan1")
        with self.subTest(msg="Global Disabled"):
            self.assertEqual(interface.proxy_arp_enabled, False)

    def test_banner(self):
        test_cases = [
            {
                "test_name": "Singleline-01",
                "config": (
                    "banner motd ^CTest Banner^C"
                ),
                "result": None
            },
            {
                "test_name": "Multiline-01",
                "config": (
                    "banner motd ^CHello\n"
                    "World\n"
                    "^C"
                ),
                "result": None
            },
            {
                "test_name": "Multiline-02",
                "config": (
                    "banner motd ^CHello\n"
                    "World\n"
                    "End^C\n"
                    "banner login ^CWoo\n"
                    "Foo^C\n"
                ),
                "result": None
            }
        ]
        for test_case in test_cases:
            with self.subTest(msg=test_case['test_name']):
                config = IosConfigParser(config=test_case['config'])
                config.parse()
                print(config.banner)
                #TODO: Make proper evaluation

    def test_password_encryption(self):
        test_cases = [
            {
                "test_name": "Default",
                "config": (
                    "!"
                ),
                "result": False

            },
            {
                "test_name": "Disabled",
                "config": (
                    "no service password-encryption"
                ),
                "result": False

            },
            {
                "test_name": "Enabled",
                "config": (
                    "service password-encryption"
                ),
                "result": True

            }

        ]
        for test_case in test_cases:
            with self.subTest(msg=test_case['test_name']):
                config = IosConfigParser(config=test_case['config'])
                config.parse()
                want = test_case['result']
                have = config.password_encryption_enabled
                print(have)
                self.assertEqual(want, have)

    def test_pad(self):
        test_cases = [
            {
                "test_name": "Default",
                "config": (
                    "!"
                ),
                "result": True

            },
            {
                "test_name": "Disabled",
                "config": (
                    "no service pad"
                ),
                "result": False

            },
            {
                "test_name": "Enabled",
                "config": (
                    "service pad"
                ),
                "result": True

            }

        ]
        for test_case in test_cases:
            with self.subTest(msg=test_case['test_name']):
                config = IosConfigParser(config=test_case['config'])
                config.parse()
                want = test_case['result']
                have = config.service_pad_enabled
                print(have)
                self.assertEqual(want, have)

    def test_ip_finger(self):
        test_cases = [
            {
                "test_name": "Default",
                "config": (
                    "!"
                ),
                "result": True

            },
            {
                "test_name": "Disabled",
                "config": (
                    "no ip finger"
                ),
                "result": False

            },
            {
                "test_name": "Enabled",
                "config": (
                    "ip finger"
                ),
                "result": True

            }

        ]
        for test_case in test_cases:
            with self.subTest(msg=test_case['test_name']):
                config = IosConfigParser(config=test_case['config'])
                config.parse()
                want = test_case['result']
                have = config.ip_finger_enabled
                print(have)
                self.assertEqual(want, have)


class TestIosAaaParser(BaseNetParserTest):

    VENDOR = "ios"

    def test_load_aaa_01(self):
        data_path, results_path = self.get_test_resources(test_name='aaa_config-01')
        config = IosConfigParser(config=data_path, verbosity=VERBOSITY)
        config.parse()

del BaseNetParserTest

if __name__ == '__main__':
    unittest.main()


import pathlib
import json
import unittest

from ruamel.yaml import YAML
yaml = YAML(typ='safe')
yaml.default_flow_style = False

TESTS_DIR = pathlib.Path(__file__).resolve().parent.absolute()
RESOURCES_DIR = TESTS_DIR.joinpath("resources")

class BaseNetParserTest(unittest.TestCase):

    RESOURCES_DIR = TESTS_DIR.joinpath("resources")

    def load_resource_text(self, path: pathlib.Path) -> str:
        data = None
        data = path.read_text()
        return data

    def load_resource_json(self, path: pathlib.Path) -> dict:
        data = None
        data = json.loads(path.read_text())
        return data

    def load_resource_yaml(self, path: pathlib.Path) -> dict:
        data = None
        data = yaml.load(path.read_text())
        return data


    def get_test_resources(self, test_name: str, vendor: str = None):
        if vendor is None:
            if hasattr(self, 'VENDOR'):
                vendor = self.VENDOR
            else:
                msg = "No Vendor Specified"
                raise ValueError(msg)
        vendor_resource_dir = RESOURCES_DIR.joinpath(vendor)
        if not vendor_resource_dir.is_dir():
            msg = f"Path '{vendor_resource_dir}' is not a directory."
            raise NotADirectoryError(msg)
        elif not vendor_resource_dir.exists():
            msg = f"Path '{vendor_resource_dir}' does not exist."
            raise FileNotFoundError(msg)

        data_path = vendor_resource_dir.joinpath("data").joinpath(f"{test_name}.txt")
        results_path = vendor_resource_dir.joinpath("results").joinpath(f"{test_name}.yml")

        for path in [data_path, results_path]:
            if not path.is_file():
                msg = f"Path '{path}' does not exist."
                raise FileNotFoundError(msg)
            elif not path.exists():
                msg = f"Path '{path}' does not exist."
                raise FileNotFoundError(msg)

        return data_path, results_path

    def load_test_resources(self, test_name: str, vendor: str = None):
        data_path, results_path = self.get_test_resources(test_name=test_name, vendor=vendor)

        resources = None
        try:
            resources = {
                "data": self.load_resource_text(data_path),
                "results": self.load_resource_yaml(results_path)
            }
        except Exception as e:
            raise

        return resources



import re
import pathlib
import unittest

from net_parser.config import BaseConfigParser, ConfigDiff, IosConfigDiff

from tests import RESOURCES_DIR

VERBOSITY = 4

class TestConfigDiff(unittest.TestCase):

    TEST_CLASS = ConfigDiff

    def test_toplevel_01(self):
        first = BaseConfigParser(config=RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('sample_config_01.txt'))
        first.parse()
        second = BaseConfigParser(
            config=[
                "interface Vlan1",
                " description Test Interface Vlan1",
                " ip address 192.0.2.1 255.255.255.0",
                "router ospf 1",
                " network 192.0.2.0 0.0.0.255 area0",
                "router bgp 65535",
                " neighbor 192.0.2.3 remote-as 65535",
                " address-family vpnv4",
                "  neighbor 192.0.2.3 activate",
                " exit-address-family"

            ]
        )
        second.parse()

        diff = ConfigDiff(first=first, second=second, verbosity=VERBOSITY, device_type='cisco_ios')
        with self.subTest(msg="Assert is propper instance"):
            self.assertIsInstance(diff, IosConfigDiff)
        print(diff)
        diff_lines = diff.difference()
        diff.print_diff(diff_lines)
        print(diff.get_update_config(diff_lines))

    def get_host_by_name(self, host_name):
        present = pathlib.Path(r'C:\SHARE\tsdr_conf\2021-03-08').joinpath(f"{host_name}.conf")
        wanted = pathlib.Path(r'C:\Users\mhudec\Develop\GitHub\tsdr_inventory_loader\config').joinpath(f"{host_name}.conf")
        present, wanted = [BaseConfigParser(config=x, verbosity=5) for x in [present, wanted]]
        [x.parse() for x in [present, wanted]]
        return present, wanted


if __name__ == '__main__':
    unittest.main()


import pathlib
import re
import unittest
from tests import BaseNetParserTest

from net_models.models.services.cisco_ios.IosLineModels import *
from net_models.models.services.cisco_ios.AaaMethods import *

from net_parser.config import (
    BaseConfigParser, IosConfigParser, BaseConfigLine, IosConfigLine,
    IosVrfDefinitionParser, IosInterfaceParser, IosLoggingLine, IosAaaParser,
    IosLineParser
)


VERBOSITY = 5


class TestIosConfigLine(BaseNetParserTest):

    VENDOR = 'ios'
    TEST_CLASS = IosConfigLine

    def test_return_types(self):
        test_cases = {
            "interface Vlan1": IosInterfaceParser,
            "vrf definition TEST": IosVrfDefinitionParser,
            "ip vrf definition TEST": IosVrfDefinitionParser,
            "logging host 192.0.2.1": IosLoggingLine,
            "aaa authentication login default local": IosAaaParser

        }
        config = IosConfigParser(config=[])
        for text, subclass in test_cases.items():
            line = self.TEST_CLASS(number=1, text=text, config=config)
            with self.subTest(msg=text):
                self.assertIsInstance(line, subclass)


class TestIosVrfDefinitionParser(BaseNetParserTest):

    VENDOR = 'ios'
    TEST_CLASS = IosVrfDefinitionParser

    def test_01(self):
        data_path, results_path = self.get_test_resources(test_name='vrf_definition_01')
        want = self.load_resource_yaml(path=results_path)
        config = IosConfigParser(config=data_path, verbosity=VERBOSITY)
        config.parse()
        have = [x.serial_dict(exclude_none=True) for x in config.vrfs]
        self.assertEqual(want, have)


class TestIosLineParser(BaseNetParserTest):

    VENDOR = 'ios'
    TEST_CLASS = IosLineParser

    def test_line_type_and_range(self):
        test_cases = [
            {
                "test_name": "Test-Console-01",
                "config": "line con 0",
                "result": ['console', [0]]
            },
            {
                "test_name": "Test-Vty-01",
                "config": "line vty 0 4",
                "result": ['vty', [0, 4]]
            },
            {
                "test_name": "Test-Aux-01",
                "config": "line aux 0",
                "result": ['aux', [0]]
            }
        ]
        config = IosConfigParser(config=[])
        config.parse()
        for test_case in test_cases:
            with self.subTest(msg=f"Type_{test_case['test_name']}"):
                line = self.TEST_CLASS(number=1, text=test_case['config'], config=config)
                want = test_case['result'][0]
                have = line.line_type
                self.assertEqual(want, have)
            with self.subTest(msg=f"Range_{test_case['test_name']}"):
                line = self.TEST_CLASS(number=1, text=test_case['config'], config=config)
                want = test_case['result'][1]
                have = line.line_range
                self.assertEqual(want, have)

    def test_exec_timeout(self):
        test_cases = [
            {
                "test_name": "Test-Exec_timeout-01",
                "config": (
                    "line con 0\n"
                    " exec-timeout 10 0"
                ),
                "result": 600
            },
            {
                "test_name": "Test-Exec_timeout-None",
                "config": (
                    "line con 0\n"
                ),
                "result": None
            },
            {
                "test_name": "Test-Exec_timeout-0",
                "config": (
                    "line con 0\n"
                    " exec-timeout 0 0"
                ),
                "result": 0
            }
        ]
        for test_case in test_cases:
            config = IosConfigParser(config=test_case['config'])
            config.parse()
            with self.subTest(msg=f"{test_case['test_name']}"):
                text_line = test_case['config'].split('\n')[0]
                line = [x for x in config.lines if x.text == text_line][0]
                want = test_case['result']
                have = line.exec_timeout
                self.assertEqual(want, have)

    def test_transport(self):
        test_cases = [
            {
                "test_name": "Test-Transport-01",
                "config": (
                    "line vty 0 4\n"
                    " transport input all\n"
                    " transport output none\n"
                    " transport preferred ssh"
                ),
                "result": IosLineTransport(input='all', output='none', preferred='ssh')
            },
            {
                "test_name": "Test-Transport-02",
                "config": (
                    "line vty 0 4\n"
                    " transport input ssh\n"
                ),
                "result": IosLineTransport(input='ssh')
            }
        ]
        for test_case in test_cases:
            config = IosConfigParser(config=test_case['config'])
            config.parse()
            with self.subTest(msg=f"{test_case['test_name']}"):
                text_line = test_case['config'].split('\n')[0]
                line = [x for x in config.lines if x.text == text_line][0]
                want = test_case['result']
                have = line.transport
                self.assertEqual(want, have)


    def test_acl(self):
        test_cases = [
            {
                "test_name": "Test-ACL-01",
                "config": (
                    "line vty 0 4\n"
                    " access-class 1 in\n"
                ),
                "result": [IosLineAccessClass(name=1, direction='in', vrf_also=False)]
            }
        ]
        for test_case in test_cases:
            config = IosConfigParser(config=test_case['config'])
            config.parse()
            with self.subTest(msg=f"{test_case['test_name']}"):
                text_line = test_case['config'].split('\n')[0]
                line = [x for x in config.lines if x.text == text_line][0]
                want = test_case['result']
                have = line.acls
                self.assertEqual(want, have)

    def test_exec_enabled(self):
        test_cases = [
            {
                "test_name": "Test-Exec-enabled-01",
                "config": (
                    "line aux 0\n"
                    " no exec\n"
                ),
                "result": False
            },
            {
                "test_name": "Test-Exec-enabled-02",
                "config": (
                    "line aux 0\n"
                ),
                "result": True
            }
        ]
        for test_case in test_cases:
            config = IosConfigParser(config=test_case['config'])
            config.parse()
            with self.subTest(msg=f"{test_case['test_name']}"):
                text_line = test_case['config'].split('\n')[0]
                line = [x for x in config.lines if x.text == text_line][0]
                want = test_case['result']
                have = line.exec_enabled
                self.assertEqual(want, have)

    def test_aaa(self):
        test_cases = [
            {
                "test_name": "Test-Exec-enabled-01",
                "config": (
                    "line vty 0 4\n"
                    " login authentication TEST\n"
                    " authorization exec TEST\n"
                    " authorization commands 15 TEST\n"
                ),
                "result": IosLineAaaConfig(
                    authentication='TEST',
                    authorization=IosAaaLineAuthorization(
                        exec='TEST',
                        commands=[
                            IosAaaLineCommands(
                                name='TEST',
                                level=15
                            )
                        ]
                    )
                )
            }
        ]
        for test_case in test_cases:
            config = IosConfigParser(config=test_case['config'])
            config.parse()
            with self.subTest(msg=f"{test_case['test_name']}"):
                text_line = test_case['config'].split('\n')[0]
                line = [x for x in config.lines if x.text == text_line][0]
                want = test_case['result']
                have = line.aaa
                print(have.yaml(exclude_none=True))
                self.assertEqual(want, have)



del BaseNetParserTest

if __name__ == '__main__':
    unittest.main()


import threading
import queue
import sys
import time
import logging


from net_parser.utils import get_logger

class BaseConnectionWorker(threading.Thread):

    def __init__(self, task_queue: queue.Queue, logger: logging.Logger = None, result_queue: queue.Queue = None, name: str = None, _name: str = "BaseConnectionWorker"):
        self.logger = logger or get_logger(name=_name)
        self.logger.debug(msg=f"Worker spawned.")
        self.shutdown_flag = threading.Event()
        self.task_queue = task_queue
        self.result_queue = result_queue
        super().__init__(name=name)

    def shutdown(self):
        self.logger.info(msg=f"Received Shutdown event - exiting.")
        self.shutdown_flag.set()

    def run(self):
        while not self.shutdown_flag.is_set():
            try:
                task = self.task_queue.get(block=True, timeout=1)
            except queue.Empty:
                self.logger.info(msg="Task Queue empty - exiting.")
                break

            if task is None:
                self.logger.warning(msg="Received None task - exiting.")
                break

            result = self.process_task(task=task)
            self.task_queue.task_done()
            self.logger.debug(msg=f"Finished task, result: {result}")
            if self.result_queue is not None:
                self.result_queue.put(result)

    def process_task(self, task):
        self.logger.debug(msg=f"Received task: {task}")
        return {"result": None}

    @classmethod
    def worker_factory(cls, task_queue: queue.Queue, result_queue: queue.Queue = None, logger: logging.Logger = None, num_workers: int = 5):
        workers = []
        for i in range(num_workers):
            worker = cls(name="WorkerThread-{}".format(i), task_queue=task_queue, result_queue=result_queue, logger=logger)
            workers.append(worker)
        return workers

    @classmethod
    def start_workers(cls, task_queue, workers):
        [w.start() for w in workers]
        while True:
            try:
                time.sleep(0.1)
                if task_queue.empty():
                    break
            except KeyboardInterrupt as e:
                [w.shutdown() for w in workers]
                sys.exit(1)
        task_queue.join()



import unittest
from tests import BaseNetParserTest

from net_parser.config import BaseConfigLine, BaseConfigParser

class TestBaseConfigParser(BaseNetParserTest):

    TEST_CLASS = BaseConfigLine
    SAMPLE_CONFIG = BaseConfigParser(config=BaseNetParserTest.RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('sample_config_01.txt'))
    SAMPLE_CONFIG.parse()

    def test_get_children(self):
        with self.subTest(msg="Interface Vlan1 - Get Children from Interface"):
            parent_line = [x for x in self.SAMPLE_CONFIG if x.text == "interface Vlan1"][0]
            children = parent_line.get_children()
            print(children)
            # self.assertEqual(description_line.parent, interface_line)

        with self.subTest(msg="Router BGP - Get only direct children"):
            parent_line = [x for x in self.SAMPLE_CONFIG if x.text == "router bgp 65535"][0]
            children = parent_line.get_children(max_depth=1)
            for c in children:
                print(c)

    def test_get_parent(self):

        with self.subTest(msg="Interface Vlan1 - Get Parent from description"):
            description_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == " description Test Interface Vlan1"][0]
            interface_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "interface Vlan1"][0]
            self.assertEqual(description_line.parent, interface_line)

        with self.subTest(msg="Interface Vlan1 - Get All Parents from description"):
            description_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == " description Test Interface Vlan1"][0]
            interface_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "interface Vlan1"][0]
            self.assertTrue(len(description_line.get_parents()) == 1)
            self.assertEqual(description_line.get_parents()[0], interface_line)

        with self.subTest(msg="Router BGP - Get All Parents from neighbor activate"):
            neighbor_activate_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "  neighbor 192.0.2.2 activate"][0]
            print(neighbor_activate_line.get_parents())
            # print(neighbor_activate_line.get_line)
            self.assertTrue(len(neighbor_activate_line.get_parents()) == 2)

    def test_get_path(self):
        with self.subTest(msg="Router BGP - Get path from neighbor activate"):
            neighbor_activate_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "  neighbor 192.0.2.2 activate"][0]
            self.assertEqual(neighbor_activate_line.get_path, ['router bgp 65535', ' address-family vpnv4'])

    def test_get_line(self):
        with self.subTest(msg="Router BGP - Get line from neighbor activate"):
            neighbor_activate_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "  neighbor 192.0.2.2 activate"][0]
            self.assertEqual(neighbor_activate_line.get_line, ['router bgp 65535', ' address-family vpnv4', '  neighbor 192.0.2.2 activate'])


    def test_get_section_by_parents(self):
        with self.subTest(msg="Router BGP - Get line from neighbor activate"):
            neighbor_activate_line = [x for x in self.SAMPLE_CONFIG.lines if x.text == "  neighbor 192.0.2.2 activate"][0]
            parent_1 = [x for x in self.SAMPLE_CONFIG if x.text == "router bgp 65535"][0]
            parent_2 = [x for x in self.SAMPLE_CONFIG if x.text == " address-family vpnv4" and x.parent == parent_1][0]
            parents = [parent_1, parent_2]
            print(parents)
            print(self.SAMPLE_CONFIG.get_section_by_parents(parents=[x.text for x in parents]))






del BaseNetParserTest

if __name__ == '__main__':
    unittest.main()


import pathlib
import re
import timeit
from typing import Union, List, Type
from net_parser.utils import get_logger, load_text, first_candidate_or_none, compile_regex, match_to_dict, property_autoparse
from net_parser.config import BaseConfigLine

from net_models.inventory import ConfigDefaults

re._MAXCACHE = 1024


class BaseConfigParser(object):

    PATTERN_TYPE = type(re.compile(pattern=""))
    CONFIG_LINE_CLS = BaseConfigLine

    def __init__(self, config: Union[pathlib.Path, List[str], str], verbosity: int = 4, name: str = "BaseConfigParser", defaults: Type[ConfigDefaults] = None, **kwargs):
        """
        Base class for parsing Cisco-like configs

        Args:
            config (:obj:`pathlib.Path` or `str` or `list`): Config file in a form of `pathlib.Path`, or `string`
                containing the entire config or list of lines of the config file
            verbosity (:obj:`int`, optional): Determines the verbosity of logging output, defaults to 4: Info

        Attributes:
            lines (list): Contains list of all config lines stored as objects (see :class:`ccutils.ccparser.BaseConfigLine`)
            config_lines_str (list): Contains list of all config lines stored as strings

        Examples:

            Possible config inputs::

                # Using pathlib
                config_file = pathlib.Path("/path/to/config_file.txt")
                config = BaseConfigParser(config=config_file)

                # Using string
                config_string = '''
                hostname RouterA
                !
                interface Ethernet0/0
                 description Test Interface
                 ip address 10.0.0.1 255.255.255.0
                !
                end
                '''
                config = BaseConfigParser(config=config_string)

                # Using list
                config_list = [
                "hostname RouterA",
                    "!",
                    "interface Ethernet0/0",
                    " description Test Interface",
                    " ip address 10.0.0.1 255.255.255.0",
                    "!",
                    "end"
                ]
                config = BaseConfigParser(config=config_list)

        """
        self.verbosity = verbosity
        self.logger = get_logger(name=name, verbosity=verbosity)
        self._config = config
        self.lines = []
        self.DEFAULTS = defaults or ConfigDefaults()


    def __iter__(self):
        return iter(self.lines)

    def __repr__(self):
        return f"[BaseConfigParser - {len(self.lines)} lines]"

    def __str__(self):
        return '\n'.join(map(lambda x: x.text, self.lines))

    def load_config(self) -> List[str]:
        raw_lines = load_text(obj=self._config, logger=self.logger)
        return raw_lines

    def parse(self):
        """
        Entry function which triggers the parsing process. Called automatically when instantiating the object.

        :return: ``None``
        """
        raw_config_lines = self.load_config()
        self.config_lines_str = raw_config_lines
        self._create_cfg_line_objects()

    def _check_path(self, filepath):
        path = None
        if not isinstance(filepath, pathlib.Path):
            path = pathlib.Path(filepath)
        else:
            path = filepath
        path = path.resolve()
        if not path.exists():
            self.logger.error(msg="Path '{}' does not exist.".format(filepath))
            return None
        if not path.is_file():
            self.logger.error(msg="Path '{}' is not a file.".format(filepath))
        if not path.is_absolute():
            path = path.resolve()
            self.logger.debug("Path '{}' is existing file.".format(filepath))
            return path
        else:
            self.logger.debug("Path '{}' is existing file.".format(filepath))
            return path

    def _get_indent(self, line):
        indent_size = len(line) - len(line.lstrip(" "))
        return indent_size

    # Deprecated
    def _get_clean_config(self, first_line_regex=r"^version \d+\.\d+", last_line_regex=r"^end"):
        self.logger.debug(msg="Cleaning config lines")
        first_regex = re.compile(pattern=first_line_regex, flags=re.MULTILINE)
        last_regex = re.compile(pattern=last_line_regex, flags=re.MULTILINE)
        first = None
        last = None
        for i in range(len(self.config_lines_str)):
            if not first:
                if re.match(pattern=first_regex, string=all_lines[i]):
                    first = i
                    self.logger.debug(msg="Found first config line: '{}'".format(all_lines[first]))
            if not last:
                if re.match(pattern=last_regex, string=all_lines[i]):
                    last = i
                    self.logger.debug(msg="Found last config line: '{}'".format(all_lines[last]))
                    break
        if not first or not last:
            self.config_lines_str = []
            self.logger.error(msg="No valid config found!")
        else:
            self.config_lines_str = all_lines[first:last + 1]
            self.logger.info(msg="Loading {} config lines.".format(len(self.config_lines_str)))
        # Fix indent

    def fix_indents(self):
        """
        Function for fixing the indentation level of config lines.

        :return:
        """
        indent_map = list(map(self._get_indent, self.config_lines_str))
        fixed_indent_map = []
        for i in range(len(indent_map)):
            if i == 0:
                ### Assume the first line is not indented
                fixed_indent_map.append(0)
                continue
            if indent_map[i] == 0:
                fixed_indent_map.append(0)
                continue
            # If indent is same preceding line, copy its indent
            if indent_map[i] == indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1])
            # If indent is higher that preceding line, increase by one
            elif indent_map[i] > indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1]+1)
            # If indent is lower that preceding line
            elif indent_map[i] < indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1]-1)
        for i, val in enumerate(fixed_indent_map):
            self.config_lines_str[i] = " "*val + self.config_lines_str[i].strip()
            #print(val, "'{}'".format(self.config_lines_str[i]))

    def _create_cfg_line_objects(self):
        """
        Function for generating ``self.lines``.

        """
        start = timeit.default_timer()
        for number, text in enumerate(self.config_lines_str):
            self.lines.append(self.CONFIG_LINE_CLS(number=number, text=text, config=self, verbosity=self.verbosity).return_obj())
        for line in self.lines:
            line.type = line.get_type
        self.logger.debug(msg="Created {} ConfigLine objects in {} ms.".format(len(self.lines), (timeit.default_timer()-start)*1000))

    def _compile_regex(self, regex, flags=re.MULTILINE):
        """
        Helper function for compiling `re` patterns from string.

        :param str regex: Regex string
        :param flags: Flags for regex pattern, default is re.MULTILINE
        :return:
        """
        pattern = None
        try:
            pattern = re.compile(pattern=regex, flags=flags)
        except Exception as e:
            self.logger.error(msg="Error while compiling regex '{}'. Exception: {}".format(regex, repr(e)))
        return pattern

    def find_objects(self, regex, flags=re.MULTILINE, group: Union[int, str, None] = None):
        """
        Function for filtering Config Lines Objects based on given regex.

        Args:
            regex (:obj:`re.Pattern` or `str`): Regex based on which the search is done
            flags (:obj:`int`, optional): Set custom flags for regex, defaults to ``re.MULTILINE``

        Examples:

            Example::

                # Initialize the object
                config = BaseConfigParser(config="/path/to/config_file.cfg")

                # Define regex for matching config lines
                interface_regex = r"^ interface"

                # Apply the filter
                interface_lines = config.find_objects(regex=interface_regex)

                # Returns subset of ``self.lines`` which match specified regex

        """
        pattern = None
        if not isinstance(regex, self.PATTERN_TYPE):
            pattern = self._compile_regex(regex=regex, flags=flags)
        else:
            pattern = regex
        lines = []
        results = []
        for line in self.lines:
            result = line.re_search(regex=pattern, group=group)
            if result:
                lines.append(line)
                results.append(result)
        if group is None:
            results = list(lines)
        else:
            return results
        self.logger.debug(msg="Matched {} lines for query '{}'".format(len(results), regex))
        return results

    def get_section_by_parents(self, parents):
        if not isinstance(parents, list):
            parents = list(parents)
        section = list(self.lines)
        for parent in parents:
            section = [x.get_children() for x in section if bool(x.is_parent and x.re_match(parent))]
            if len(section) == 1:
                section = section[0]
            elif len(section) > 1:
                self.logger.error("Multiple lines matched parent statement. Cannot determine config section.")
                return []
            else:
                self.logger.error("No lines matched parent statement. Cannot determine config section.")
                return []
        return section

    def match_to_dict(self, line, patterns):
        return match_to_dict(line=line, regexes=patterns)

    def property_autoparse(self, candidate_pattern, patterns):
        return property_autoparse(lines=self.lines, candidate_pattern=candidate_pattern, regexes=patterns, logger=self.logger, include_candidate=True)

    def section_property_autoparse(self, parent, patterns, return_with_line=False):
        entries = None
        if isinstance(parent, BaseConfigLine):
            candidates = [parent]
        elif isinstance(parent, (str, self.PATTERN_TYPE)):
            candidates = self.find_objects(regex=parent)
        if len(candidates):
            entries = []
        else:
            return entries
        for candidate in candidates:
            entry = {}
            if isinstance(parent, (str, self.PATTERN_TYPE)):
                entry.update(self.match_to_dict(line=candidate, patterns=[parent]))
            for pattern in patterns:
                updates = candidate.re_search_children(regex=pattern, group="ALL")
                if len(updates) == 1:
                    entry.update(updates[0])
                elif len(updates) == 0:
                    entry.update({k: None for k in pattern.groupindex.keys()})
                else:
                    self.logger.warning("Multiple possible updates found for Pattern: '{}' on Candidate: '{}'".format(pattern, candidate))
            if return_with_line:
                entries.append((candidate, entry))
            else:
                entries.append(entry)
        return entries

    def compile_regex(self, pattern: str, flags=re.MULTILINE):
        return compile_regex(pattern=pattern, flags=flags, logger=self.logger, raise_exc=True)

    def _val_to_bool(self, entry: dict, keys: list):
        if not isinstance(keys, list):
            keys = list(keys)
        for key in keys:
            if entry[key]:
                entry[key] = True
            else:
                entry[key] = False
        return entry

    def first_candidate_or_none(self, candidates: list, wanted_type=None):
        return first_candidate_or_none(candidates=candidates, logger=self.logger, wanted_type=wanted_type)



import pathlib
import unittest
from net_parser.utils.common import (check_path, load_text, re_search_lines, raw_match_lines)
from net_parser.utils.get_logger import get_logger
from net_parser.exceptions import *

from net_parser.config import BaseConfigParser

from tests import RESOURCES_DIR, BaseNetParserTest

TEST_LOGGER = get_logger(name="TEST-LOGGER", verbosity=5)

class TestCheckPath(unittest.TestCase):


    def test_invalid_path_01(self):
        with self.assertRaises(expected_exception=FileNotFoundError):
            check_path(path=r"This is not a valid path.txt", logger=TEST_LOGGER)

    def test_invalid_path_02(self):
        with self.assertRaises(expected_exception=FileNotFoundError):
            check_path(path=r"\nonexistent.txt", logger=TEST_LOGGER)

    def test_invalid_path_03(self):
        with self.assertRaises(expected_exception=(InvalidPathSyntax, FileNotFoundError)):
            check_path(
                path="This is not\n"
                     "a\n"
                     "valid path",
                logger=TEST_LOGGER
            )

    def test_valid_path_01(self):
        path = check_path(
            path=RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('test_load_01.txt'),
            logger=TEST_LOGGER
        )
        self.assertIsInstance(path, pathlib.Path)

    def test_valid_path_02(self):
        path = check_path(
            path=str(RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('test_load_01.txt')),
            logger=TEST_LOGGER
        )
        self.assertIsInstance(path, pathlib.Path)


class TestLoadText(unittest.TestCase):

    def test_load_singleline(self):
        lines = load_text(obj="Singleline string", logger=TEST_LOGGER)
        self.assertIsInstance(lines, list)
        self.assertEqual(len(lines), 1)

    def test_load_multiline(self):
        lines = load_text(obj="Multiline\nstring", logger=TEST_LOGGER)
        self.assertIsInstance(lines, list)
        self.assertEqual(len(lines), 2)

    def test_load_valid_list(self):
        test_lines = [
            "interface Loopback0",
            " description Test"
        ]
        lines = load_text(obj=test_lines, logger=TEST_LOGGER)
        self.assertIsInstance(lines, list)
        self.assertEqual(len(lines), 2)

    def test_load_invalid_list(self):
        test_lines = [
            "interface Loopback0",
            0,
            None
        ]
        with self.assertRaises(expected_exception=AssertionError):
            lines = load_text(obj=test_lines, logger=TEST_LOGGER)


    def test_load_valid_path(self):
        lines = load_text(
            obj=RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('test_load_01.txt'),
            logger=TEST_LOGGER
        )
        self.assertIsInstance(lines, list)

    def test_load_valid_path_str(self):
        lines = load_text(
            obj=str(RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('test_load_01.txt')),
            logger=TEST_LOGGER
        )
        self.assertIsInstance(lines, list)
        self.assertTrue(len(lines) > 0)

    def test_load_invalid_path(self):
        with self.assertRaises(expected_exception=FileNotFoundError):
            lines = load_text(
                obj=pathlib.Path("nonexistent\path"),
                logger=TEST_LOGGER
            )



class CommonUtilsTest(unittest.TestCase):

    SAMPLE_CONFIG = BaseConfigParser(config=RESOURCES_DIR.joinpath('ios').joinpath('data').joinpath('sample_config_01.txt'))
    SAMPLE_CONFIG.parse()


class TestReSearchLines(CommonUtilsTest):

    def test_re_search_lines(self):

        lines = re_search_lines(lines=self.SAMPLE_CONFIG.lines, regex=r"interface Vlan1")
        want = [x for x in self.SAMPLE_CONFIG.lines if x.text == "interface Vlan1"][0]
        have = lines[0]
        self.assertTrue(want is have)


class TestRawMatchLines(CommonUtilsTest):

    def test_raw_match_lines(self):

        lines = raw_match_lines(lines=self.SAMPLE_CONFIG.lines, text=r"interface Vlan1")
        want = "interface Vlan1"
        have = lines[0].text
        self.assertEqual(want, have)


if __name__ == '__main__':
    unittest.main()


import functools
import logging
import pathlib
import re
from pydantic.typing import Union, List, Tuple, Type

from net_parser.exceptions import *
from .get_logger import get_logger

PATTERN_TYPE = type(re.compile(pattern=""))

LOGGER = get_logger(name="CommonLogger", verbosity=4)

EMPTY_LINE_REGEX = re.compile(pattern=r'^\s*$')
NOT_EMPTY_LINE_REGEX = re.compile(pattern=r'.*\S.*')

def check_path(path: pathlib.Path, logger: logging.Logger) -> pathlib.Path:
    if not isinstance(path, pathlib.Path):
        try:
            path = pathlib.Path(path)
        except Exception as e:
            msg = f"Unhandled Exception occured while converting string to path. Exception: {repr(e)}"
            logger.critical(msg=msg)
            raise
    try:
        path = path.resolve()
    except OSError as e:
        msg = "Path syntax is invalid"
        logger.debug(msg=msg)
        raise InvalidPathSyntax(msg)
    except Exception as e:
        msg = f"Unhandled Exception occured while trying to resolve path. Exception: {repr(e)}"
        logger.critical(msg=msg)
        raise
    # If we got here, the path syntax is valid, just check if it exists and is file
    if not path.is_file():
        msg = f"File '{path}' not found."
        # logger.error(msg=msg)
        raise FileNotFoundError(msg)

    return path


def load_text(obj: Union[pathlib.Path, List[str], str], logger: logging.Logger = None, omit_empty_lines: bool = False) -> List[str]:
    logger = logger or LOGGER
    lines = []
    path = None
    # Decide base on type of obj:
    if isinstance(obj, list):
        # Loading object is list
        if not all([isinstance(x, str) for x in obj]):
            msg = "Expected list of str, but not all elements are strings."
            logger.error(msg=msg)
            raise AssertionError(msg)
        else:
            lines = list(obj)
    elif isinstance(obj, str):
        # Might be a path
        try:
            path = check_path(path=obj, logger=logger)
            lines = path.read_text().splitlines()
        except FileNotFoundError as e:
            path = None
            lines = obj.splitlines()
        except InvalidPathSyntax as e:
            path = None
            lines = obj.splitlines()

    elif isinstance(obj, pathlib.Path):
        try:
            path = check_path(path=obj, logger=logger)
            lines = path.read_text().splitlines()
        except FileNotFoundError as e:
            msg = f"Got path to load, but the path does not exist. Path: {obj}"
            logger.critical(msg=msg)
            raise
    all_lines = len(lines)
    lines = [x.rstrip() for x in lines if not EMPTY_LINE_REGEX.match(x)]
    empty_lines = all_lines - len(lines)
    logger.debug(f"Loaded {len(lines)} lines. {empty_lines} were empty.")
    return lines


def first_candidate_or_none(candidates: list, logger: logging.Logger = None, wanted_type=None):
    logger = logger or LOGGER
    if len(candidates) == 0:
        return None
    elif len(candidates) == 1:
        if wanted_type is None:
            return candidates[0]
        else:
            return wanted_type(candidates[0])
    else:
        logger.error(msg='Multiple candidates found.')
        return None

@functools.lru_cache(maxsize=512)
def compile_regex(pattern: str, logger: logging.Logger = None, flags=re.MULTILINE, raise_exc=True) -> Type[PATTERN_TYPE]:
    """
    Wrapper around re.compile with some defaults and error handling.

    Args:
        pattern: Regular Expression string.
        logger: Instance of logging.Logger, if not provided, default logger will be used.
        flags: Flags for the re.compile, uses `re.MULTILINE` by default.
        raise_exc: Wether or not to raise exception of re.compile fails. Default is True.

    Returns: A compiled pattern

    """
    logger = logger or LOGGER
    compiled_regex = None
    try:
        compiled_regex = re.compile(pattern=pattern, flags=flags)
    except Exception as e:
        logger.error(msg=f"Error while compiling regex '{compiled_regex}'. Exception: {repr(e)}")
        if raise_exc:
            raise
        else:
            compiled_regex = None
    return compiled_regex


def assert_is_regex(regex) -> Type[PATTERN_TYPE]:
    """
    Try to make sure that `regex` is actually a re.compile-d patter. If not, make it so.

    Args:
        regex: Expects either compiled patter or string.

    Returns: Compiled pattern

    """
    if isinstance(regex, PATTERN_TYPE):
        return regex
    else:
        return compile_regex(pattern=regex, raise_exc=True)

def re_search(line: 'BaseConfigLine', regex: re.Pattern, group: str = None):
    """
    Search config line for given regex

    Args:
        regex (:obj:`re.Pattern` or ``str``): Regex to search for
        group (:obj:`str` or ``int``, optional): Return only specific (named or numbered) group of given regex.
            If set to "ALL", return value will be a dictionary with all named groups of the regex.

    Examples:

        Example::

            # Given the following line stored in `line` variable
            # " ip address 10.0.0.1 255.255.255"
            pattern = r"^ ip address (?P<ip>\S+) (?P<mask>\S+)"

            # Basic search
            result = line.re_search(regex=pattern)
            print(result)
            # Returns: " ip address 10.0.0.1 255.255.255"

            # Search for specific group
            result = line.re_search(regex=pattern, group="ip")
            print(result)
            # Returns: "10.0.0.1"

            # Get all named groups
            result = line.re_search(regex=pattern, group="ALL")
            print(result)
            # Returns: {"ip": "10.0.0.1", "mask": "255.255.255"}


    Returns:
        str: String that matched given regex, or, if `group` was provided, returns only specific group.

        Returns ``None`` if regex did not match.

    """
    logger = line.logger
    regex = assert_is_regex(regex)

    m = regex.search(string=line.text)
    if m:
        if group is None:
            return m.group(0)
        elif isinstance(group, int):
            if group <= regex.groups:
                return m.group(group)
            else:
                logger.error(msg="Given regex '{}' does not contain required group '{}'".format(regex, group))
                return None
        elif isinstance(group, str):
            if group in regex.groupindex.keys():
                return m.group(group)
            elif group == "ALL":
                return m.groupdict()

            else:
                logger.error(msg="Given regex '{}' does not contain required group '{}'".format(regex, group))
                return None
    else:
        logger.debug(msg="Given regex '{}' did not match.".format(regex))
        return None

def match_to_dict(line: Type['BaseConfigLine'], regexes: List[Type[PATTERN_TYPE]]) -> dict:
    """
    Match multiple regexes against sigle line and return a dictionary with all the named groups
    contained in all regexes.

    Args:
        line: Instance of BaseConfigLine and its subclasses
        regexes: List of regexes to match against the line

    Returns: A flat dictionary where keys are named groups from the regexes.

    """
    entry = {}
    for regex in regexes:
        regex = assert_is_regex(regex)
        match_result = line.re_search(regex=regex, group="ALL")
        if match_result is not None:
            entry.update(match_result)
        else:
            entry.update({k: None for k in regex.groupindex.keys()})
    return entry



def re_search_lines(lines: List['BaseConfigLine'], regex: re.Pattern, group: str = None) -> List[Union[str, dict, Type['BaseConfigLine']]]:
    regex = assert_is_regex(regex)
    result = list(filter(lambda x: bool(regex.search(string=x.text)), lines))
    if group is not None:
        result = [x.re_search(regex=regex, group=group) for x in result]
    return result

def raw_match_lines(lines: List[Type['BaseConfigLine']], text: str) -> List[Type['BaseConfigLine']]:
    results = [x for x in lines if x.text == text]
    return results

def re_filter_lines(lines: List['BaseConfigLine'], regex: re.Pattern, group: str = None, pop_matches: bool = True) -> Tuple[list, list]:
    regex = assert_is_regex(regex)
    result = re_search_lines(lines=lines, regex=regex, group=None)
    if pop_matches:
        lines = sorted(list(set(lines).difference(set(result))), key=lambda x: x.number)
    if group is not None:
        result = [x.re_search(regex=regex, group=group) for x in result]
    return result, lines

def re_match(line: 'BaseConfigLine', regex, group=None) -> List[Union[str, dict, Type['BaseConfigLine']]]:
    logger = line.logger
    regex = assert_is_regex(regex)

    if regex is None:
        return None
    m = regex.match(string=line.text)
    if m:
        if group is None:
            return m.group(0)
        elif isinstance(group, int):
            if group <= regex.groups:
                return m.group(group)
            else:
                logger.error(msg="Given regex '{}' does not contain required group '{}'".format(regex, group))
                return None
        elif isinstance(group, str):
            if group in regex.groupindex.keys():
                return m.group(group)
            elif group == "ALL":
                return m.groupdict()
            else:
                logger.error(msg="Given regex '{}' does not contain required group '{}'".format(regex, group))
                return None
    else:
        logger.debug(msg="Given regex '{}' did not match.".format(regex))
        return None


def property_autoparse(lines: List['BaseConfigLine'], candidate_pattern: re.Pattern, regexes: List[re.Pattern], logger: logging.Logger = None, include_candidate: bool = True):
    logger = logger or LOGGER
    properties = None
    candidates = re_search_lines(lines=lines, regex=candidate_pattern)
    if len(candidates):
        properties = []
    else:
        return properties
    if include_candidate and candidate_pattern not in regexes:
        regexes.insert(0, candidate_pattern)
    for candidate in candidates:
        properties.append(match_to_dict(line=candidate, regexes=regexes))
    return properties


import logging
import sys
import pathlib

LOG_PATH = pathlib.Path.home().joinpath(".net_parser.log")
VERBOSITY_MAP = {
    1: logging.CRITICAL,
    2: logging.ERROR,
    3: logging.WARNING,
    4: logging.INFO,
    5: logging.DEBUG
}
def get_logger(name, verbosity=4, handle=["stderr"], with_threads=False) -> logging.Logger:
    """
    This function provides common logging facility by creating instances of `loggers` from python standard ``logging`` library.
    :param str name: Name of the logger
    :param bool DEBUG: Enables/disables debugging output
    :param list handle: Changing value of this parameter is not recommended.
    :return: Instance of logger object
    """
    global VERBOSITY_MAP
    threading_formatter_string = '[%(asctime)s] [%(levelname)s]\t[%(name)s][%(threadName)s][%(module)s][%(funcName)s]\t%(message)s'
    single_formatter_string = '[%(asctime)s] [%(levelname)s]\t[%(name)s][%(module)s][%(funcName)s]\t%(message)s'

    formatter_string = threading_formatter_string if with_threads else single_formatter_string

    logfile_path = LOG_PATH
    formatter = logging.Formatter(formatter_string)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stderr_handler = logging.StreamHandler(sys.stderr)
    file_handler = logging.FileHandler(logfile_path, delay=True)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)

    handlers = []
    if "stderr" in handle:
        handlers.append(stderr_handler)
    if "stdout" in handle:
        handlers.append(stdout_handler)


    for handler in handlers:
        handler.setFormatter(formatter)
        try:
            handler.setLevel(VERBOSITY_MAP[verbosity])
        except KeyError:
            handler.setLevel(logging.INFO)

    root = logging.getLogger(name)
    root.propagate = 0
    root.setLevel(logging.DEBUG)
    if verbosity == 0:
        root.disabled = True
    has_handler = {"file_handler": False, "stderr_handler": False, "stdout_handler": False}
    for handler in root.handlers:
        if isinstance(handler, logging.FileHandler):
            has_handler["file_handler"] = True
        if isinstance(handler, logging.StreamHandler):
            has_handler["stderr_handler"] = True
    if not has_handler["file_handler"]:
        root.addHandler(file_handler)
    if not has_handler["stderr_handler"]:
        root.addHandler(stderr_handler)

    return root


import functools
import re
import pathlib
import timeit
from typing import Union, List, Generator, Type

from net_models.validators import normalize_interface_name
from net_models.models.interfaces.InterfaceModels import InterfaceModel
from net_models.models import VRFModel
from net_models.models.services.ServerModels import *
from net_models.inventory import HostConfig, ConfigDefaults


from net_parser.utils import re_search_lines, re_filter_lines, compile_regex, property_autoparse
from net_parser.config import (
    BaseConfigParser, BaseConfigLine, IosConfigLine,
    IosConfigParser, IosInterfaceParser, IosAaaParser, IosVrfDefinitionParser, IosLineParser, IosLoggingLine
)


class IosConfigParser(BaseConfigParser):

    INTERFACE_LINE_CLASS = IosInterfaceParser
    CONFIG_LINE_CLS = IosConfigLine

    _interface_pattern = r"[A-z]{2,}(?:[A-z\-])?\d+(?:\/\d+)?(?:\:\d+)?(?:\.\d+)?"
    _ip_address_pattern = r"(?:\d{1,3}\.){3}\d{1,3}"
    _host_pattern = r"[A-z0-9\-\_\.]+"
    _source_interface_regex = re.compile(pattern=r"source (?P<src_interface>{0})".format(_interface_pattern))
    _source_vrf_regex = re.compile(pattern=r"vrf (?P<vrf>\S+)")

    _hostname_regex = re.compile(pattern=r"^hostname (?P<hostname>\S+)\Z")
    _ip_arp_proxy_disable_regex = re.compile(pattern=r"^(?:(?P<no>no) )?ip arp proxy disable$", flags=re.MULTILINE)
    _service_password_encryption_regex = re.compile(pattern=r"^(?:(?P<no>no) )?service password-encryption$", flags=re.MULTILINE)
    _banner_regex = re.compile(pattern=r"^banner (?P<banner_type>\S+)")
    _service_pad_regex = re.compile(pattern=r"^(?:(?P<no>no) )?service pad$", flags=re.MULTILINE)
    _ip_finger_regex = re.compile(pattern=r"^(?:(?P<no>no) )?ip finger(?: (?P<rfc_compliant>rfc-compliant))?$", flags=re.MULTILINE)
    _ip_source_route_regex = re.compile(pattern=r"^(?:(?P<no>no) )?ip source-route$", flags=re.MULTILINE)

    _ntp_server_base_regex = re.compile(pattern=r"^ntp server(?: vrf \S+)? (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern), flags=re.MULTILINE)
    _ntp_peer_base_regex = re.compile(pattern=r"^ntp peer(?: vrf \S+)? (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern), flags=re.MULTILINE)
    _ntp_authentication_keys_regex = re.compile(pattern=r"^ntp authentication-key (?P<key_id>\d+) (?P<method>\S+) (?P<value>\S+)(?: (?P<encryption_type>\d+))?", flags=re.MULTILINE)
    _ntp_trusted_key_regex = re.compile(pattern=r"^ntp trusted-key (?P<key_id>\d+)", flags=re.MULTILINE)
    _ntp_acl_regex = re.compile(pattern=r"^ntp access-group (?P<access_type>\S+) (?P<acl_name>\S+)", flags=re.MULTILINE)
    _ntp_src_interface_regex = re.compile(pattern=r"^ntp source (?P<src_interface>{})".format(_interface_pattern), flags=re.MULTILINE)

    _logging_source_interface_regex = re.compile(pattern=r"^logging source-interface (?P<src_interface>{0})(?: vrf (?P<vrf>\S+))?".format(_interface_pattern))
    _logging_server_base_regex = re.compile(pattern=r"^logging host (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern))
    _logging_transport_regex = re.compile(pattern=r"transport (?P<protocol>udp|tcp) port (?P<port>\d+)")

    def __init__(self,
                 config: Union[pathlib.Path, List[str], str],
                 verbosity: int =4,
                 name: str = "BaseConfigParser",
                 defaults: Type[ConfigDefaults] = None,
                 **kwargs):
        super().__init__(config=config, verbosity=verbosity, name="IosConfigParser", **kwargs)
        self.DEFAULTS = defaults or ConfigDefaults()

    @functools.cached_property
    def hostname(self):
        candidates = self.find_objects(regex=self._hostname_regex, group="hostname")
        return self.first_candidate_or_none(candidates=candidates)

    @property
    def interface_lines(self) -> Generator[IosInterfaceParser, None, None]:
        return (x for x in self.lines if 'interface' in x.get_type)

    @property
    def interfaces(self) -> Generator[InterfaceModel, None, None]:
        return (x.to_model() for x in self.interface_lines)

    @functools.cached_property
    def aaa_lines(self):
        return (x for x in self.lines if isinstance(x, IosAaaParser))

    @functools.cached_property
    def management_lines(self):
        return (x.to_model() for x in self.lines if isinstance(x, IosLineParser))

    @functools.cached_property
    def logging_lines(self):
        return (x for x in self.lines if isinstance(x, IosLoggingLine))

    def get_interface_line(self, interface_name: str) -> Union[IosInterfaceParser, None]:
        interface_name = normalize_interface_name(interface_name=interface_name, short=False)
        candidates = [x for x in self.interface_lines if x.name == interface_name]
        return self.first_candidate_or_none(candidates=candidates)

    @property
    def vrf_definition_lines(self) -> Generator[IosVrfDefinitionParser, None, None]:
        return (x for x in self.lines if isinstance(x, IosVrfDefinitionParser))

    @property
    def vrfs(self) -> Generator[VRFModel, None, None]:
        return (x.model for x in self.vrf_definition_lines)

    @functools.cached_property
    def ntp(self) -> NtpConfig:
        ntp = NtpConfig()
        ntp_lines = self.find_objects(regex=self.compile_regex(pattern=r"^ntp .*", flags=re.MULTILINE))
        # Source Interface
        ntp_src_interface = self.first_candidate_or_none(candidates=re_search_lines(lines=ntp_lines, regex=self._ntp_src_interface_regex, group="src_interface"))
        if ntp_src_interface is not None:
            ntp.src_interface = ntp_src_interface
        # Servers Section
        candidate_pattern = self._ntp_server_base_regex
        regexes = [
            self._source_interface_regex,
            self._source_vrf_regex,
            re.compile("key (?P<key_id>\d+)"),
            re.compile("(?P<prefer>prefer)")
        ]

        ntp_servers = property_autoparse(lines=ntp_lines, candidate_pattern=candidate_pattern, regexes=regexes, logger=self.logger, include_candidate=True)
        ntp_servers = [self._val_to_bool(entry=x, keys=['prefer']) for x in ntp_servers]
        ntp_servers = [NtpServer.parse_obj(x) for x in ntp_servers]
        if len(ntp_servers):
            ntp.servers = ntp_servers
        # Peers
        candidate_pattern = self._ntp_peer_base_regex
        regexes = [
            self._source_interface_regex,
            self._source_vrf_regex,
            re.compile("key (?P<key_id>\d+)"),
            re.compile("(?P<prefer>prefer)")
        ]
        ntp_peers = property_autoparse(lines=ntp_lines, candidate_pattern=candidate_pattern, regexes=regexes, logger=self.logger, include_candidate=True)
        ntp_peers = [self._val_to_bool(entry=x, keys=['prefer']) for x in ntp_peers]
        ntp_peers = [NtpServer.parse_obj(x) for x in ntp_peers]
        if len(ntp_peers):
            ntp.peers = ntp_peers
        authenticate = self._globals_check(
            regex=re.compile(
                pattern=r"^(?:(?P<no>no) )?ntp authenticate$",
                flags=re.MULTILINE
            ),
            default=False
        )
        ntp.authenticate = authenticate
        # Keys
        ntp_auth_keys, ntp_lines = re_filter_lines(lines=ntp_lines, regex=self._ntp_authentication_keys_regex, group='ALL')
        ntp_auth_keys = [NtpKey.parse_obj(x) for x in ntp_auth_keys]
        ntp_trusted_keys = [int(x) for x in re_search_lines(lines=ntp_lines, regex=self._ntp_trusted_key_regex, group='key_id')]
        for ntp_key in ntp_auth_keys:
            if ntp_key.key_id in ntp_trusted_keys:
                ntp_key.trusted = True
        if len(ntp_auth_keys):
            ntp.ntp_keys = ntp_auth_keys
        # Access Lists
        acl_lines, ntp_lines = re_filter_lines(lines=ntp_lines, regex=self._ntp_acl_regex, group="ALL")
        if len(acl_lines):
            ntp.access_groups = NtpAccessGroups()
        for entry in acl_lines:
            if entry['access_type'] == 'serve-only':
                ntp.access_groups.serve_only = entry['acl_name']
            if entry['access_type'] == 'query-only':
                ntp.access_groups.query_only = entry['acl_name']
            if entry['access_type'] == 'serve':
                ntp.access_groups.serve = entry['acl_name']
            if entry['access_type'] == 'peer':
                ntp.access_groups.peer = entry['acl_name']
        return ntp

    @functools.cached_property
    def logging(self) -> LoggingConfig:
        logging_lines = list(self.logging_lines)
        if len(logging_lines) == 0:
            return None
        logging = LoggingConfig()
        candidate_pattern = self._logging_server_base_regex
        regexes = [
            self._source_vrf_regex,
            self._source_interface_regex,
            self._logging_transport_regex
        ]
        logging_servers = property_autoparse(lines=logging_lines, candidate_pattern=self._logging_server_base_regex, regexes=regexes, logger=self.logger, include_candidate=True)
        logging_servers = [{k:v for k,v in x.items() if v is not None} for x in logging_servers]
        logging_servers = [LoggingServer.parse_obj(x) for x in logging_servers]
        if len(logging_servers):
            logging.servers = logging_servers
        logging_sources = re_search_lines(lines=logging_lines, regex=self._logging_source_interface_regex, group='ALL')
        logging_sources = [LoggingSource.parse_obj(x) for x in logging_sources]
        if len(logging_sources):
            logging.sources = logging_sources
        return logging




    @property
    def routing(self):
        raise NotImplementedError

    @functools.cached_property
    def proxy_arp_enabled(self) -> bool:
        candidates = self.find_objects(regex=self._ip_arp_proxy_disable_regex, group='ALL')
        candidate = self.first_candidate_or_none(candidates=candidates)
        if candidate is not None:
            candidate = self._val_to_bool(entry=candidate, keys=['no'])
            if candidate['no'] is True:
                # no ip arp proxy disable
                return True
            elif candidate['no'] is False:
                # ip arp proxy disable
                return False
        else:
            # Enabled by default
            return True

    def _globals_check(self, regex: re.Pattern, default: bool) -> bool:
        """
        This function looks for given `regex`, if found and prefixed with 'no', returns False. If not prefixed with no,
        returns True. Returns 'default' otherwise.
        Args:
            regex: re.Pattern with 'no' groups specified
            default: bool - which value to return if regex did not match

        Returns:

        """
        candidates = self.find_objects(regex=regex, group='ALL')
        candidate = self.first_candidate_or_none(candidates=candidates)
        if candidate is not None:
            candidate = self._val_to_bool(entry=candidate, keys=['no'])
            if candidate['no'] is True:
                # Negated
                return False
            elif candidate['no'] is False:
                return True
        else:
            return default

    @functools.cached_property
    def password_encryption_enabled(self) -> bool:
        return self._globals_check(regex=self._service_password_encryption_regex, default=False)

    @functools.cached_property
    def banner(self):
        banners = {}
        candidates = self.find_objects(regex=self._banner_regex)
        stop_chars = ['^C', chr(3)]
        for candidate in candidates:
            banner_type = candidate.re_search(regex=self._banner_regex, group='banner_type')
            banner_text = None
            # Determine the stopchar
            stop_char_occurences = {candidate.text.count(x):x for x in stop_chars}
            stop_char = stop_char_occurences[max(stop_char_occurences.keys())]
            if max(stop_char_occurences.keys()) == 2: # SingleLine
                banner_text = [x for x in candidate.text.split(stop_char) if x != ''][-1]
            else: # Multiline
                banner_text = []
                # First line
                first_part_candidates = [x for x in candidate.text.split(stop_char) if x != ''][1:]
                if len(first_part_candidates):
                    banner_text.append(first_part_candidates[0])
                for line in self.lines[candidate.number+1:]:
                    if stop_char in line.text:
                        last_part_candidate = [x for x in line.text.split(stop_char) if x != ''][:1]
                        if len(last_part_candidate):
                            banner_text.append(last_part_candidate[0])
                        break
                    else:
                        banner_text.append(line.text)
            if isinstance(banner_text, list):
                banner_text = '\n'.join(banner_text)
            banners[banner_type] = banner_text
        return banners


    @functools.cached_property
    def ip_source_routing_enabled(self):
        return self._globals_check(regex=self._ip_source_route_regex, default=True)

    @functools.cached_property
    def ip_finger_enabled(self):
        return self._globals_check(regex=self._ip_finger_regex, default=True)


    @functools.cached_property
    def service_pad_enabled(self) -> bool:
        """
        Packet Assembler Disassembler service

        Returns: bool

        """
        return self._globals_check(regex=self._service_pad_regex, default=True)


    @functools.cached_property
    def service_tcp_keepalives_in(self):
        self._globals_check(regex=self._service_tcp_keepalives_in, default=False)

    @functools.cached_property
    def service_tcp_keepalives_in(self):
        self._globals_check(regex=self._service_tcp_keepalives_in, default=False)


    def to_model(self):
        model = HostConfig(interfaces={x.name: x for x in self.interfaces})

        vrfs = list(self.vrfs)
        if len(vrfs):
            model.vrf_definitions = vrfs

        management_lines = list(self.management_lines)
        if len(management_lines):
            model.management_lines = management_lines

        if self.ntp is not None:
            model.ntp = self.ntp

        if self.logging is not None:
            model.logging = self.logging

        return model

    def __repr__(self):
        return f"[IosConfigParser - {len(self.lines)} lines]"



import functools
import re

from net_parser.utils import get_logger
from typing import (Union, List, Dict)

class BaseOpsParser(object):

    logger = get_logger(name="BaseOpsParser", verbosity=4)
    VENDOR = None
    COMANDS = []
    PATTER_MAP = {}

    @classmethod
    def update_class_logger(cls, verbosity: int = 4):
        cls.logger = get_logger(name=cls.__name__, verbosity=verbosity)

    @classmethod
    def max_rounds(cls):
        return max(cls.PATTER_MAP.keys())

    @classmethod
    @functools.lru_cache()
    def _get_group_names(cls, pattern: re.Pattern):
        return [x for x in pattern.groupindex.keys() if isinstance(x, str)]

    @classmethod
    def build_entry(cls, match: re.Match):
        entry = None
        if len(match.groupdict().keys()):
            entry = match.groupdict()
        else:
            entry = match.group(0)
        return entry

    @classmethod
    def update_dict(cls, orig: dict, new: dict, overwrite: bool = False):
        result = dict(orig)
        if overwrite:
            result.update({k:v for k, v in new.items() if v is not None})
        else:
            for k, v in new.items():
                if result.get(k) is None:
                    result[k] = v
        return result

    @classmethod
    def match_single_pattern(cls, string: str, pattern: re.Pattern):
        cls.logger.debug(msg=f"Got {pattern.pattern=} to match {string=}")
        group_names = cls._get_group_names(pattern=pattern)
        for m in re.finditer(pattern=pattern, string=string):
            if len(group_names):
                if len(group_names) == 1 and group_names[0] == '_entry':
                    yield m.group('_entry')
                else:
                    yield m.groupdict()
            else:
                yield m.group(0)

    @classmethod
    def match_multi_pattern(cls, text: str, patterns: List[re.Pattern]):
        entry = {}
        for pattern in patterns:
            entry.update({k:None for k in cls._get_group_names(pattern=pattern)})
        # Populate Named Groups
        for pattern in patterns:
            match = pattern.search(string=text)
            if match:
                entry = cls.update_dict(orig=entry, new=match.groupdict())
                if all(entry.values()):
                    cls.logger.debug(msg=f"All fields matched, stopping search.")
                    break
        cls.logger.debug(msg=f"Returning {entry=}")
        return entry


    def __str__(cls):
        return f"[{cls.__class__.__name__}]"

    def __repr__(cls):
        return cls.__str__()


class IosOpsParser(BaseOpsParser):

    VENDOR = 'ios'


class IosCdpNeighborDetailParser(IosOpsParser):

    COMANDS = ['show cdp neighbors detail']
    PATTER_MAP = {
        1: {
            "_entries": [
                re.compile(
                    pattern=r"^(-{10,})\n(?P<_entry>Device ID:(?:.*(?:\r\n|\r|\n))*?(?=\1|\Z))",
                    flags=re.MULTILINE
                )
            ]
        },
        2: {
            "_entry": [
                re.compile(
                    pattern=r"Device ID: (?P<device_id>\S+)",
                    flags=re.MULTILINE
                )
            ]
        },
        3: {
            "device_id": [
                re.compile(
                    pattern="^(?P<hostname>.*?)(?:\.(?P<domain_name>\S+))?$",
                    flags=re.MULTILINE
                )
            ]
        }
    }









from pydantic.typing import (
    Dict, List, Literal, Optional, Union, Type
)

from net_parser.utils import get_logger, first_candidate_or_none, raw_match_lines, PATTERN_TYPE, assert_is_regex
from net_parser.config import BaseConfigParser, BaseConfigLine


__all__ = [
    'DiffedLine',
    'ConfigDiff',
    'ConfigDiff',
    'IosConfigDiff'
]

class DiffedLine:

    def __init__(self, text, action: Literal['add', 'remove', 'present']):
        self.text = text
        self.action = action

    def __str__(self):
        return self.text

    def __repr__(self):
        sign_map = {
            'add': '+',
            'remove': '-',
            'present': '='
        }
        return f"{sign_map[self.action]} {self.text}"


class ConfigDiff:
    _registry = {}

    def __init_subclass__(cls, **kwargs):
        device_type = kwargs.pop('device_type', None)
        if device_type is not None:
            cls._registry[device_type] = cls
        super().__init_subclass__(**kwargs)

    def __new__(cls, *args, **kwargs):
        device_type = kwargs.pop('device_type', None)
        subclass = None
        if device_type is not None:
            subclass = cls._registry.get(device_type, None)
        if subclass is None:
            subclass = cls
        obj = object.__new__(subclass)
        obj.__init__(**kwargs)
        return obj


    DEFAULT_IGNORE_LINES = []

    def __init__(self,
                 first: Type[BaseConfigParser],
                 second: Type[BaseConfigParser],
                 verbosity: int = 4,
                 **kwargs):
        self.logger = get_logger(name='ConfigDiff', verbosity=verbosity)
        self.first = first
        self.second = second
        self._check_configs()

    def _check_configs(self):
        for i, config in enumerate([self.first, self.second]):
            if not isinstance(config, BaseConfigParser):
                msg = f"Config {'first' if i == 0 else 'second'} has to be an instance of 'BaseConfigParser', got {type(self.first)=}"
                self.logger.critical(msg=msg)
                raise TypeError(msg)

    def add_line(self, line):
        return DiffedLine(text=line.text, action='add')

    def remove_line(self, line):
        return DiffedLine(text=line.text, action='remove')

    def present_line(self, line):
        return DiffedLine(text=line.text, action='present')

    def is_ignore_line(self, line: Type[BaseConfigLine], diff_ignore: List[Type[PATTERN_TYPE]] = None) -> bool:
        ignore = False
        diff_ignore = diff_ignore or self.DEFAULT_IGNORE_LINES
        if not isinstance(diff_ignore, list):
            diff_ignore = [diff_ignore]
        diff_ignore = [assert_is_regex(x) for x in diff_ignore]
        if len(diff_ignore) == 0:
            return False
        else:
            ignore = any([bool(x.match(string=line.text)) for x in diff_ignore])
        if ignore:
            self.logger.info(msg=f"Ignored line: {line}")
        return ignore

    def diff_line(self, line: Type[BaseConfigLine], other_lines: List[Type[BaseConfigLine]], diff_ignore: List[Type[PATTERN_TYPE]] = None):
        if 'comment' in line.get_type:
            return []
        if self.is_ignore_line(line=line, diff_ignore=diff_ignore):
            return []
        updates = []

        candidates = raw_match_lines(lines=other_lines, text=line.text)
        first_candidate = first_candidate_or_none(candidates=candidates[:1])
        if first_candidate is None:
            updates.append(self.add_line(line))
            if line.is_parent:
                updates.extend([self.add_line(x) for x in line.get_children()])
        elif line.is_parent and not first_candidate.is_parent:
            updates.append(self.present_line(first_candidate))
            # Add all children to update
            updates.extend([self.add_line(x) for x in line.get_children()])
        elif line.is_parent and first_candidate.is_parent:
            line_updates = []
            for child in line.get_children(max_depth=1):
                if self.is_ignore_line(line=child, diff_ignore=diff_ignore):
                    continue
                # For each direct child generate diff and append to updates
                child_updates = self.diff_line(line=child, other_lines=first_candidate.get_children())
                if len(child_updates):
                    line_updates.extend(child_updates)
            if len(line_updates):
                line_updates.insert(0, self.present_line(line))
            updates.extend(line_updates)
        self.logger.debug(msg=f"Returning updates:\n{updates}")
        return updates

    def difference(self, diff_ignore: List[Type[PATTERN_TYPE]] = None):
        diff_updates = []
        first_top_lines = [x for x in self.first.lines if x.indent == 0]
        second_top_lines = [x for x in self.second.lines if x.indent == 0]
        for line in second_top_lines:
            line_diff = self.diff_line(line=line, other_lines=first_top_lines, diff_ignore=diff_ignore)
            diff_updates.extend(line_diff)
        return diff_updates

    @staticmethod
    def print_diff(diff_lines: List[DiffedLine]):
        present = len([x for x in diff_lines if x.action == 'present'])
        added = len([x for x in diff_lines if x.action == 'add'])
        removed = len([x for x in diff_lines if x.action == 'remove'])
        print(f"! Diff:\n!\t{present=}\n!\t{added=}\n!\t{removed=}")
        for line in diff_lines:
            print(repr(line))

    @staticmethod
    def get_update_config(diff_lines: List[DiffedLine]):
        return '\n'.join([x.text for x in diff_lines if x.action in ['present', 'add']])

    def __str__(self):
        return f"[{self.__class__.__name__} Object]"


class IosConfigDiff(ConfigDiff, device_type='cisco_ios'):

    pass



import pathlib
import re
import json
import timeit
import functools


from net_parser.utils import get_logger, first_candidate_or_none, re_search, re_search_lines, re_match, PATTERN_TYPE, compile_regex


class BaseConfigLine(object):

    PATTERN_TYPE = PATTERN_TYPE
    _parent_indent_regex = re.compile(pattern=r"[^! ]", flags=re.MULTILINE)
    _interface_regex = re.compile(pattern=r"^interface\s(\S+)", flags=re.MULTILINE)
    comment_regex = re.compile(pattern=r"^\s*!.*", flags=re.MULTILINE)

    def __init__(self, number, text, config: 'BaseConfigParser', verbosity=3, name="BaseConfigLine"):
        """
        **This class is not meant to be instantiated directly, but only from BaseConfigParser instance.**

        Args:
            number (int): Index of line in config
            text (str): Text of the config line
            config (:obj:`BaseConfigParser`): Reference to the parent BaseConfigParser object
            verbosity (:obj:`int`, optional): Logging output level, defaults to 3: Warning

        """
        self._name = name
        self.logger = get_logger(name=name, verbosity=verbosity)
        #print(self.logger.handlers)
        self.config = config
        self.config_lines_obj = self.config.lines
        self.number = number
        self.text = text
        self.indent = len(self.text) - len(self.text.lstrip(" "))
        self.type = None
        # self.logger.debug("Parsing line: #{}: '{}'".format(self.number, self.text))


    def return_obj(self):
        return self

    # def __eq__(self, other) -> bool:
    #     return self.get_line == other.get_line

    # def __ne__(self, other) -> bool:
    #     return not self.__eq__(other)

    def compile_regex(self, pattern: str, flags=re.MULTILINE):
        return compile_regex(pattern=pattern, flags=flags, logger=self.logger, raise_exc=True)

    def get_children(self, max_depth: int = None):
        """
        Return all children lines (all following lines with larger indent)

        Returns:
            list: List of child config lines (objects)

        """
        children = []
        line_num = int(self.number) + 1
        #print(len(self.config.lines))
        while line_num <= len(self.config.lines) - 1:   # Avoid IndexError
            if self.config.lines[line_num].indent <= self.indent:
                break
            else:
                if max_depth and self.config.lines[line_num].indent > self.indent + max_depth:
                    pass
                else:
                    children.append(self.config.lines[line_num])
                line_num += 1
        return children

    @property
    def parent(self):
        if not self.is_child:
            self.logger.debug("Line is not a child, therefore has no parent. Line: {}".format(self.text))
            return None
        else:
            line_num = int(self.number) - 1
            line = self.config.lines[line_num]
            while line.indent >= self.indent and line_num > 0:
                line_num -= 1
                line = self.config.lines[line_num]
            return line

    @functools.lru_cache()
    def get_parents(self):
        start = timeit.default_timer()
        parents = []
        if not self.is_child:
            self.logger.debug("Line is not a child, therefore has no parent. Line: {}".format(self.text))
            pass
        else:
            parents.insert(0, self.parent)
            while parents[0].parent is not None:
                parents.insert(0, parents[0].parent)
        stop = timeit.default_timer()
        self.logger.debug("Getting parents of line {} took {} ms".format(str(self), (stop-start)*10e3))
        return parents

    @property
    def get_path(self):
        path = [x.text for x in self.get_parents()]
        if len(path):
            return path
        else:
            return None

    @property
    def get_line(self):
        line = None
        if self.get_path is not None:
            line = list(self.get_path)
        else:
            line = list()
        line.append(self.text)
        return line


    def re_search_children(self, regex, group=None):
        """
        Search all children for given regex.

        Args:
            regex (:obj:`re.Pattern` or ``str``): Regex to search for
            group (:obj:`str` or ``int``, optional): Return only specific (named or numbered) group of given regex.
                If set to "ALL", return value will be a dictionary with all named groups of the regex.

        Returns:
            list: List of all child object which match given regex, or, if `group` was provided, returns
            list containing matched group across all children.

                Example::

                    # Given following config section, interface line stored in `line` variable
                    config = '''
                    interface Ethernet0/0
                     description Test Interface
                     ip address 10.0.0.1 255.255.255.0
                     ip address 10.0.1.1 255.255.255.0 secondary
                    !
                    '''
                    pattern = r"^ ip address (?P<ip>\S+) (?P<mask>\S+)"

                    result = line.re_search_children(regex=pattern)
                    print(result)
                    # Returns: [
                    #   [BaseConfigLine #2 (child): ip address 10.0.0.1 255.255.255.0],
                    #   [BaseConfigLine #3 (child): ip address 10.0.1.1 255.255.255.0 secondary]
                    # ]

                    result = line.re_search_children(regex=pattern, group="ip")
                    print(result)
                    # Returns: [
                    #   "10.0.0.1",
                    #   "10.0.1.1"
                    # ]

                    result = line.re_search_children(regex=pattern, group="ALL")
                    print(result)
                    # Returns: [
                    #   {"ip": "10.0.0.1", "mask": "255.255.255.0"},
                    #   {"ip": "10.0.1.1", "mask": "255.255.255.0"}
                    # ]


        """
        pattern = None
        if not isinstance(regex, self.PATTERN_TYPE):
            pattern = self.compile_regex(pattern=regex)
        else:
            pattern = regex
        if not pattern:
            return []
        children = self.get_children()
        return re_search_lines(lines=children, regex=regex, group=group)

    # TODO: Add Tests
    # TODO: Add Examples
    def re_search_children_multipattern(self, regexes: list, group=None, deduplicate: bool = True) -> list:
        """
        Wrapper function for ``self.re_search_children()`` allowing to use multiple patterns

        Args:
            regexes (``list``): List of patterns to search
            group (``str`` or ``int``, optional): Return only specific (named or numbered) group of given regex.
                If set to "ALL", return value will be a dictionary with all named groups of the regex.
            deduplicate (``bool``, optional): When set to ``True`` (default), results will not contain duplicate line objects in
                cases where multiple patterns match the same line.

        Returns:
            list: List of all child object which match given regex, or, if `group` was provided, returns
            list containing matched group across all children.

        """
        results = []
        for regex in regexes:
            for result in [x for x in self.re_search_children(regex=regex, group=group)]:
                if result in results:
                    if deduplicate:
                        continue
                    else:
                        results.append(result)
                else:
                    results.append(result)
        return results

    def re_search(self, regex, group=None):
        return re_search(line=self, regex=regex, group=group)

    def re_match(self, regex, group=None):
        return re_match(line=self, regex=regex, group=group)

    @property
    def get_type(self):
        """
        Return `types` of config line. Used mostly for filtering purposes.

        Currently available values are:

        - ``parent``
        - ``child``
        - ``interface``
        - ``comment``

        Returns:
            list: List of types

        """
        types = []
        if re.match(self.comment_regex, self.text):
            types.append("comment")
            # If line is comment, it's comment only
            return types
        if self.is_parent:
            types.append("parent")
        if self.is_child:
            types.append("child")
        return types

    @functools.cached_property
    def is_comment(self):
        return bool(re.match(self.comment_regex, self.text))

    @property
    def is_parent(self):
        """
        Check whether this line is a parent

        Returns:
            bool: True if line is a parent line, False otherwise

        """
        if self.number < len(self.config.lines)-1:
            if self.config.lines[self.number+1].indent > self.indent:
                return True
            else:
                return False
        else:
            return False

    @property
    def is_child(self):
        """
           Check whether this line is a child

           Returns:
               bool: True if line is a child line, False otherwise

           """
        if self.indent > 0:
            return True
        else:
            return False

    def _val_to_bool(self, entry: dict, keys: list):
        if not isinstance(keys, list):
            keys = list(keys)
        for key in keys:
            if entry[key]:
                entry[key] = True
            else:
                entry[key] = False
        return entry

    def first_candidate_or_none(self, candidates: list, wanted_type=None):
        return first_candidate_or_none(candidates=candidates, logger=self.logger, wanted_type=wanted_type)

    def __str__(self):
        return f"[{self._name} #{self.number}\t({self.get_type}): '{self.text}']"

    def __repr__(self):
        return self.__str__()





import functools
import re
from pydantic.typing import (
    List,
    Union,
    Literal
)


from net_models.models.BaseModels.SharedModels import VRFAddressFamily, VRFModel
from net_models.models.services.cisco_ios.IosLineModels import *
from net_models.models.services.cisco_ios.AaaMethods import *

from net_parser.utils import re_search
from net_parser.config import BaseConfigLine

AAA_SECTION_REGEX = re.compile(pattern=r'^aaa \S+.*$', flags=re.MULTILINE)
VRF_SECTION_REGEX = re.compile(pattern=r'^(?:ip )?vrf definition \S+.*$', flags=re.MULTILINE)
LOGGING_SECTION_REGEX = re.compile(pattern=r'^logging \S+.*$', flags=re.MULTILINE)
LINE_SECTION_REGEX = re.compile(pattern=r'^line .*$')



class IosConfigLine(BaseConfigLine):
    _registry = {}

    comment_regex = re.compile(pattern=r"^\s*!.*", flags=re.MULTILINE)

    def __init_subclass__(cls, regex: re.Pattern = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if regex is not None:
            cls._registry[regex] = cls

    def __new__(cls, *args, **kwargs):
        text = kwargs.get('text')
        subclass = None
        for pattern, subclass_candidate in cls._registry.items():
            # print(f"Testing {pattern=} against {text=}")
            if pattern.match(string=text):
                # print("Pattern matched")
                subclass = subclass_candidate
                break
        if subclass is None:
            subclass = cls
        instance = object.__new__(subclass)
        instance.__init__(*args, **kwargs)
        return instance

    def __init__(self, number: int, text: str, config, verbosity: int = 4, name: str = "IosConfigLine"):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name=name)




class IosAaaParser(IosConfigLine, regex=AAA_SECTION_REGEX):

    def __init__(self, number: int, text: str, config: 'IosConfigParser', verbosity: int = 4):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name="IosAaaLine")


class IosVrfDefinitionParser(IosConfigLine, regex=VRF_SECTION_REGEX):

    _name_regex = re.compile(pattern=r"^(?:ip )?vrf definition (?P<name>\S+)", flags=re.MULTILINE)
    _description_regex = re.compile(pattern=r"^ description (?P<description>.*?)\Z", flags=re.MULTILINE)
    _rd_regex = re.compile(pattern=r"^ rd (?P<rd>\S+)\Z", flags=re.MULTILINE)
    _address_family_regex = re.compile(pattern=r"^ address-family (?P<afi>\S+)(?: (?P<safi>\S+))?\Z")
    _route_target_regex = re.compile(pattern=r"^  route-target (?P<action>import|export) (?P<rt>\S+)(?: (?P<rt_type>\S+))?", flags=re.MULTILINE)

    def __init__(self, number: int, text: str, config, verbosity: int = 4):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name="IosVrfDefinitionLine")

    @property
    def get_type(self):
        types = super().get_type
        types.append('vrf')
        return types

    @property
    def name(self) -> Union[str, None]:
        return self.re_match(regex=self._name_regex, group=1)

    @property
    def description(self) -> Union[str, None]:
        candidates = self.re_search_children(regex=self._description_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates)

    @property
    def rd(self) -> bool:
        candidates = self.re_search_children(regex=self._rd_regex, group=1)
        return self.first_candidate_or_none(candidates=candidates)

    @property
    def address_families(self) -> Union[List[VRFAddressFamily], None]:
        address_families = []
        af_lines = self.re_search_children(regex=self._address_family_regex)
        for af_line in af_lines:
            data = {}
            data.update(af_line.re_search(regex=self._address_family_regex, group="ALL"))
            # Route Targets
            rt_candidates = af_line.re_search_children(regex=self._route_target_regex, group="ALL")
            print(rt_candidates)
            if len(rt_candidates):
                data['route_targets'] = rt_candidates
            if not any(data.values()):
                continue
            else:
                model = VRFAddressFamily(**data)
                address_families.append(model)
        if len(address_families):
            self.logger.debug(f"Found {len(address_families)} AFs for VRF {self.name}")
            return address_families
        else:
            self.logger.debug(f"Found no AFs for VRF {self.name}")
            return None

    @property
    @functools.lru_cache()
    def model(self):
        data = {
            'name': self.name,
            'rd': self.rd,
            'description': self.description,
            'address_families': self.address_families,
        }
        model = VRFModel(**{k:v for k, v in data.items() if v is not None})
        return model



class IosLoggingLine(IosConfigLine, regex=LOGGING_SECTION_REGEX):

    def __init__(self, number: int, text: str, config, verbosity: int = 4):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name="IosLoggingLine")

    @property
    def get_type(self):
        types = super().get_type
        types.append('logging')
        return types


class IosLineParser(IosConfigLine, regex=LINE_SECTION_REGEX):

    _line_base_regex = re.compile(pattern=r"^line (?P<line_type>\S+) (?P<range_start>\d+)(?: (?P<range_stop>\d+))?", flags=re.MULTILINE)
    _exec_timeout_regex = re.compile(pattern=r"^ exec-timeout (?P<minutes>\d+) (?P<seconds>\d+)")
    _transport_regex = re.compile(pattern=r"^ transport (?P<direction>input|output|preferred) (?P<protocol>\S+)$")
    _acl_regex = re.compile(pattern=r"^ access-class (?P<name>\S+) (?P<direction>in|out)(?: (?P<vrf_also>vrf-also)|(?: vrf (?P<vrf>\S+)))?", flags=re.MULTILINE)
    _acl_ipv6_regex = re.compile(pattern=r"^ ipv6 access-class (?P<name>\S+) (?P<direction>in|out)(?: (?P<vrf_also>vrf-also)|(?: vrf (?P<vrf>\S+)))?", flags=re.MULTILINE)
    _exec_disabled_regex = re.compile(pattern=r"^ no exec", flags=re.MULTILINE)
    _authentication_regex = re.compile(pattern=r"^ login authentication (?P<authentication>\S+)")
    _authorization_exec_regex = re.compile(pattern=r"^ authorization exec (?P<authorization>\S+)")
    _authorization_commands_regex = re.compile(pattern=r"^ authorization commands (?P<level>\d+) (?P<authorization>\S+)")
    _priv_level_regex = re.compile(pattern=r"^ privilege level (?P<priv_level>\d+)$", flags=re.MULTILINE)
    _session_timeout_regex = re.compile(pattern=r"^ session-timeout (?P<minutes>\d+)(?: +(?P<output>output))?", flags=re.MULTILINE)



    def __init__(self, number: int, text: str, config, verbosity: int = 4):
        super().__init__(number=number, text=text, config=config, verbosity=verbosity, name="IosLineLine")

    @functools.cached_property
    def get_type(self):
        types = super().get_type
        types.append('management_line')
        return types

    @functools.cached_property
    def line_type(self) -> Literal['aux', 'console', 'vty']:
        candidate = self.re_search(regex=self._line_base_regex, group='line_type')
        if candidate == 'con':
            candidate = 'console'
        return candidate

    @functools.cached_property
    def line_range(self) -> List[int]:
        range_start = self.re_search(regex=self._line_base_regex, group='range_start')
        range_stop = self.re_search(regex=self._line_base_regex, group='range_stop')
        line_range = [range_start]
        if range_stop is not None:
            line_range.append(range_stop)
        return [int(x) for x in line_range]

    @functools.cached_property
    def exec_timeout(self) -> Union[int, None]:
        candidate = self.first_candidate_or_none(
            candidates=self.re_search_children(regex=self._exec_timeout_regex, group='ALL')
        )
        if candidate is None:
            return None
        else:
            return int(candidate['minutes'])*60 + int(candidate['seconds'])

    @functools.cached_property
    def transport(self) -> IosLineTransport:
        candidates = self.re_search_children(regex=self._transport_regex, group='ALL')
        data = {}
        for candidate in candidates:
            data[candidate['direction']] = candidate['protocol']
        model = IosLineTransport.parse_obj(data)
        return model

    @functools.cached_property
    def has_ssh_enabled(self) -> bool:
        transport = self.transport
        if transport.input in ['all', 'ssh']:
            return True
        else:
            return False

    @functools.cached_property
    def has_telnet_enabled(self) -> bool:
        transport = self.transport
        if transport.input in ['all', 'telnet']:
            return True
        else:
            return False

    def has_output_enabled(self):
        transport = self.transport
        if transport.output in ['none']:
            return False
        else:
            return True

    @functools.cached_property
    def acls(self) -> List[IosLineAccessClass]:
        candidates = self.re_search_children(regex=self._acl_regex, group="ALL")
        if len(candidates) == 0:
            return None
        candidates = [self._val_to_bool(entry=x, keys=['vrf_also']) for x in candidates]
        acls = []
        for candidate in candidates:
            acl = IosLineAccessClass.parse_obj(candidate)
            acls.append(acl)
        return acls

    @functools.cached_property
    def acls_ipv6(self) -> List[IosLineAccessClass]:
        candidates = self.re_search_children(regex=self._acl_ipv6_regex, group="ALL")
        if len(candidates) == 0:
            return None
        candidates = [self._val_to_bool(entry=x, keys=['vrf_also']) for x in candidates]
        acls = []
        for candidate in candidates:
            acl = IosLineAccessClass.parse_obj(candidate)
            acls.append(acl)
        return acls

    @functools.cached_property
    def acl_inbound(self) -> IosLineAccessClass:
        acls = self.acls
        return [x for x in acls if x.direction == 'in']

    @functools.cached_property
    def acl_outbound(self) -> IosLineAccessClass:
        acls = self.acls
        return [x for x in acls if x.direction == 'out']

    @functools.cached_property
    def exec_enabled(self):
        candidates = self.re_search_children(regex=self._exec_disabled_regex)
        if len(candidates):
            return False
        else:
            return True

    @functools.cached_property
    def priv_level(self) -> int:
        candidates = self.re_search_children(regex=self._priv_level_regex, group='priv_level')
        return self.first_candidate_or_none(candidates=candidates, wanted_type=int)

    @functools.cached_property
    def session_timeout(self) -> dict:
        #TODO: Map to model
        candidates = self.re_search_children(regex=self._session_timeout_regex, group='ALL')
        candidates = [self._val_to_bool(entry=x, keys=['output'])]
        return self.first_candidate_or_none(candidates=candidates)

    @functools.cached_property
    def aaa(self) -> IosLineAaaConfig:
        # TODO: Add accounting
        # Authentication
        authentication_candidates = self.re_search_children(regex=self._authentication_regex, group='authentication')
        authentication = self.first_candidate_or_none(candidates=authentication_candidates)
        if authentication is None:
            authentication = 'default'
        # Authorization
        authorization_exec_candidates = self.re_search_children(regex=self._authorization_exec_regex, group='authorization')
        authorization_exec = self.first_candidate_or_none(candidates=authorization_exec_candidates)
        if authorization_exec is None:
            authorization_exec = 'default'
        authorization = IosAaaLineAuthorization(exec=authorization_exec)
        authorization_commands_candidates = self.re_search_children(regex=self._authorization_commands_regex, group='ALL')
        if len(authorization_commands_candidates):
            authorization.commands = []
        for c in authorization_commands_candidates:
            authorization.commands.append(IosAaaLineCommands(name=c['authorization'], level=c['level']))

        model = IosLineAaaConfig(
            authentication=authentication,
            authorization=authorization
        )
        return model

    def to_model(self):
        model = IosLineConfig(
            line_type=self.line_type,
            line_range=self.line_range,
        )


from .BaseConfigLine import BaseConfigLine
from .BaseConfigParser import BaseConfigParser
from .IosInterfaceParser import IosInterfaceParser
from .IosSectionParsers import *
from .IosConfigParser import IosConfigParser
from .ConfigDiff import *



from setuptools import setup, find_packages
import pathlib


VERSION = "0.2.0"

def load_requirements():
    requirements = []
    REQS_PATH = pathlib.Path(__file__).resolve().parent.joinpath('requirements.txt')
    if REQS_PATH.exists() and REQS_PATH.is_file():
        requirements = [x for x in REQS_PATH.read_text().splitlines() if (len(x) and not x.startswith("#"))]
    return requirements


setup(
    name="net_parser",
    packages=find_packages(),
    version=VERSION,
    author="Miroslav Hudec <http://github.com/mihudec>",
    description="Network Config Parser",
    install_requires=load_requirements(),
    include_package_data=True
)


import queue
import unittest

from net_parser.utils import get_logger
from net_parser.config import IosConfigParser
from net_parser.connections import BaseConnectionWorker

try:
    from netmiko import ConnectHandler
except ImportError:
    print("Netmiko not installed. Please install via 'pip install netmiko'")

class IosConnectionWorker(BaseConnectionWorker):

    def process_task(self, task):

        with ConnectHandler(**task["handler"]) as device:
            device.enable()
            output = device.send_command('show running-config')
            config = IosConfigParser(config=output, verbosity=5)
            config.DEFAULTS.INTERFACES_DEFAULT_NO_SHUTDOWN = True
            config.DEFAULTS.INTERFACES_DEFAULT_CDP_ENABLED = True
            config.DEFAULTS.INTERFACES_DEFAULT_LLDP_ENABLED = True
            config.parse()
            result = config.to_model().serial_dict(exclude_none=True)
            return result

class BaseConnectionTest(unittest.TestCase):

    def test_01(self):
        task_queue = queue.Queue()
        task_queue.put({
            "host_name": "P01-R1",
            "handler": {
                "ip": "10.17.84.211",
                "device_type": "cisco_ios_telnet",
                "username": "admin",
                "password": "cisco"
            }
        })

        task_queue.put({
            "host_name": "P01-R2",
            "handler": {
                "ip": "10.17.84.212",
                "device_type": "cisco_ios_telnet",
                "username": "admin",
                "password": "cisco"
            }
        })
        result_queue = queue.Queue()
        logger = get_logger(name="TEST-LOGGER", with_threads=True, verbosity=5)
        workers = IosConnectionWorker.worker_factory(task_queue=task_queue, result_queue=result_queue, logger=logger, num_workers=5)
        IosConnectionWorker.start_workers(task_queue=task_queue, workers=workers)
        print(result_queue.qsize())
        results = []
        while True:
            result = None
            try:
                result = result_queue.get_nowait()
                results.append(result)
            except queue.Empty:
                print("Result queue empty")
                break
        for result in results:
            print(result)

if __name__ == '__main__':
    unittest.main()


import unittest

from net_parser.ops.BaseOpsParsers import IosCdpNeighborDetailParser

from tests import BaseNetParserTest


class TestBaseOpsParser(BaseNetParserTest):

    pass


class TestCdpNeighborDetailParser(BaseNetParserTest):

    pass

del BaseNetParserTest

if __name__ == '__main__':
    unittest.main()